import pytest


@pytest.fixture(scope="session", autouse=True)
def ic():
    """Make `ic` available everywhere during testing for easier debugging"""
    try:
        import icecream
    except ImportError:
        return
    icecream.install()
    # icecream.ic.disable()  # do ic.enable() to re-enable
    return icecream.ic



import itertools

from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph

__all__ = ["edge_boundary", "node_boundary"]


def edge_boundary(G, nbunch1, nbunch2=None, data=False, keys=False, default=None):
    # TODO: figure out data, keys, and default arguments and handle multigraph
    # data=True will be tested in test_mst.py
    is_multigraph = G.is_multigraph()
    # This may be wrong for multi-attributed graphs
    if data is True:
        weight = "weight"
    elif not data:
        weight = None
    else:
        weight = data
    G = to_graph(G, weight=weight)
    v1 = G.set_to_vector(nbunch1, ignore_extra=True)
    v2 = G.set_to_vector(nbunch2, ignore_extra=True)
    result = algorithms.edge_boundary(G, v1, v2, is_weighted=is_multigraph or data)
    rows, cols, vals = result.to_coo(values=is_multigraph or data)
    id_to_key = G.id_to_key
    if data:
        it = zip(
            (id_to_key[row] for row in rows),
            (id_to_key[col] for col in cols),
            # Unsure about this; data argument may mean *all* edge attributes
            ({weight: val} for val in vals),
        )
    else:
        it = zip(
            (id_to_key[row] for row in rows),
            (id_to_key[col] for col in cols),
        )
    if is_multigraph:
        # Edge weights indicate number of times to repeat edges
        it = itertools.chain.from_iterable(itertools.starmap(itertools.repeat, zip(it, vals)))
    return it


def node_boundary(G, nbunch1, nbunch2=None):
    G = to_graph(G)
    v1 = G.set_to_vector(nbunch1, ignore_extra=True)
    if nbunch2 is not None:
        mask = G.set_to_vector(nbunch2, ignore_extra=True).S
    else:
        mask = None
    result = algorithms.node_boundary(G, v1, mask=mask)
    return G.vector_to_nodeset(result)



import networkx as nx

from graphblas_algorithms import DiGraph, nxapi


def test_directed():
    # XXX" is transitivity supposed to work on directed graphs like this?
    G = nx.complete_graph(5, create_using=nx.DiGraph())
    G.remove_edge(1, 2)
    G.remove_edge(2, 3)
    G.add_node(5)
    G2 = DiGraph.from_networkx(G)
    expected = nx.transitivity(G)
    result = nxapi.transitivity(G2)
    assert expected == result
    # clustering
    expected = nx.clustering(G)
    result = nxapi.clustering(G2)
    assert result == expected
    expected = nx.clustering(G, [0, 1, 2])
    result = nxapi.clustering(G2, [0, 1, 2])
    assert result == expected
    for i in range(6):
        assert nx.clustering(G, i) == nxapi.clustering(G2, i)
    # average_clustering
    expected = nx.average_clustering(G)
    result = nxapi.average_clustering(G2)
    assert result == expected
    expected = nx.average_clustering(G, [0, 1, 2])
    result = nxapi.average_clustering(G2, [0, 1, 2])
    assert result == expected
    expected = nx.average_clustering(G, count_zeros=False)
    result = nxapi.average_clustering(G2, count_zeros=False)
    assert result == expected



__all__ = ["mutual_weight"]


def mutual_weight(G, u, v):
    key_to_id = G._key_to_id
    if u not in key_to_id or v not in key_to_id:
        return 0
    u = key_to_id[u]
    v = key_to_id[v]
    A = G._A
    return A.get(u, v, 0) + A.get(v, u, 0)



from graphblas import binary, monoid, unary


def normalize(x, how):
    how = how.lower()
    if how == "l1":
        denom = x.reduce().get(0)
    elif how == "l2":
        denom = (x @ x).get(0) ** 0.5
    elif how == "linf":
        denom = x.reduce(monoid.max).get(0)
    else:
        raise ValueError(f"Unknown normalization method: {how}")
    try:
        x *= 1.0 / denom
    except ZeroDivisionError:  # pragma: no cover
        pass
    return x


def is_converged(xprev, x, tol):
    """Check convergence, L1 norm: err = sum(abs(xprev - x)); err < N * tol

    This modifies `xprev`.
    """
    xprev << binary.minus(xprev | x)
    xprev << unary.abs(xprev)
    err = xprev.reduce().get(0)
    return err < xprev.size * tol



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph

from ..exception import PowerIterationFailedConvergence

_all = ["pagerank"]


def pagerank(
    G,
    alpha=0.85,
    personalization=None,
    max_iter=100,
    tol=1e-06,
    nstart=None,
    weight="weight",
    dangling=None,
):
    G = to_graph(G, weight=weight, dtype=float)
    N = len(G)
    if N == 0:
        return {}
    # We'll normalize initial, personalization, and dangling vectors later
    x = G.dict_to_vector(nstart, dtype=float, name="nstart")
    p = G.dict_to_vector(personalization, dtype=float, name="personalization")
    row_degrees = G.get_property("plus_rowwise+")  # XXX: What about self-edges?
    if dangling is not None and row_degrees.nvals < N:
        dangling_weights = G.dict_to_vector(dangling, dtype=float, name="dangling")
    else:
        dangling_weights = None
    try:
        result = algorithms.pagerank(
            G,
            alpha=alpha,
            personalization=p,
            max_iter=max_iter,
            tol=tol,
            nstart=x,
            dangling=dangling_weights,
            row_degrees=row_degrees,
        )
        return G.vector_to_nodemap(result, fillvalue=0.0)
    except algorithms.exceptions.ConvergenceFailure as e:
        raise PowerIterationFailedConvergence(*e.args) from e



from graphblas import Matrix, monoid, replace, select, semiring

from graphblas_algorithms.classes.graph import Graph

__all__ = ["k_truss"]


def k_truss(G: Graph, k) -> Graph:
    # Ignore self-edges
    S = G.get_property("offdiag")

    if k < 3:
        # Most implementations consider k < 3 invalid,
        # but networkx leaves the graph unchanged
        C = S
    else:
        # Remove edges not in k-truss
        nvals_last = S.nvals
        # TODO: choose dtype based on max number of triangles
        plus_pair = semiring.plus_pair["int32"]
        C = Matrix("int32", S.nrows, S.ncols)
        while True:
            C(S.S, replace) << plus_pair(S @ S.T)
            C << select.value(C >= k - 2)
            if C.nvals == nvals_last:
                break
            nvals_last = C.nvals
            S = C

    # Remove isolate nodes
    indices, _ = C.reduce_rowwise(monoid.any).to_coo()
    Ktruss = C[indices, indices].new()

    # Convert back to networkx graph with correct node ids
    keys = G.list_to_keys(indices)
    key_to_id = dict(zip(keys, range(len(indices))))
    return Graph(Ktruss, key_to_id=key_to_id)



try:
    import networkx as nx
except ImportError:

    class NetworkXError(Exception):
        pass

    class NetworkXPointlessConcept(Exception):
        pass

    class NodeNotFound(Exception):
        pass

    class PowerIterationFailedConvergence(Exception):
        pass

else:
    from networkx import (
        NetworkXError,
        NetworkXPointlessConcept,
        NodeNotFound,
        PowerIterationFailedConvergence,
    )
try:
    import scipy as sp
except ImportError:

    class ArpackNoConvergence(Exception):
        def __init__(self, msg, eigenvalues, eigenvectors):
            super().__init__(msg)
            self.eigenvalues = eigenvalues
            self.eigenvectors = eigenvectors

else:
    from scipy.sparse.linalg import ArpackNoConvergence



from . import nxapi

#######
# NOTE: Remember to update README.md when adding or removing algorithms from Dispatcher
#######


class Dispatcher:
    # Boundary
    edge_boundary = nxapi.boundary.edge_boundary
    node_boundary = nxapi.boundary.node_boundary
    # Centrality
    degree_centrality = nxapi.centrality.degree_alg.degree_centrality
    eigenvector_centrality = nxapi.centrality.eigenvector.eigenvector_centrality
    in_degree_centrality = nxapi.centrality.degree_alg.in_degree_centrality
    katz_centrality = nxapi.centrality.katz.katz_centrality
    out_degree_centrality = nxapi.centrality.degree_alg.out_degree_centrality
    # Cluster
    average_clustering = nxapi.cluster.average_clustering
    clustering = nxapi.cluster.clustering
    generalized_degree = nxapi.cluster.generalized_degree
    square_clustering = nxapi.cluster.square_clustering
    transitivity = nxapi.cluster.transitivity
    triangles = nxapi.cluster.triangles
    # Community
    inter_community_edges = nxapi.community.quality.inter_community_edges
    intra_community_edges = nxapi.community.quality.intra_community_edges
    # Core
    k_truss = nxapi.core.k_truss
    # Cuts
    boundary_expansion = nxapi.cuts.boundary_expansion
    conductance = nxapi.cuts.conductance
    cut_size = nxapi.cuts.cut_size
    edge_expansion = nxapi.cuts.edge_expansion
    mixing_expansion = nxapi.cuts.mixing_expansion
    node_expansion = nxapi.cuts.node_expansion
    normalized_cut_size = nxapi.cuts.normalized_cut_size
    volume = nxapi.cuts.volume
    # DAG
    ancestors = nxapi.dag.ancestors
    descendants = nxapi.dag.descendants
    # Dominating
    is_dominating_set = nxapi.dominating.is_dominating_set
    # Isolate
    is_isolate = nxapi.isolate.is_isolate
    isolates = nxapi.isolate.isolates
    number_of_isolates = nxapi.isolate.number_of_isolates
    # Link Analysis
    hits = nxapi.link_analysis.hits_alg.hits
    pagerank = nxapi.link_analysis.pagerank_alg.pagerank
    # Reciprocity
    overall_reciprocity = nxapi.overall_reciprocity
    reciprocity = nxapi.reciprocity
    # Regular
    is_k_regular = nxapi.regular.is_k_regular
    is_regular = nxapi.regular.is_regular
    # Shortest Paths
    has_path = nxapi.shortest_paths.generic.has_path
    # Simple Paths
    is_simple_path = nxapi.simple_paths.is_simple_path
    # S Metric
    s_metric = nxapi.smetric.s_metric
    # Structural Holes
    mutual_weight = nxapi.structuralholes.mutual_weight
    # Tournament
    is_tournament = nxapi.tournament.is_tournament
    score_sequence = nxapi.tournament.score_sequence
    tournament_matrix = nxapi.tournament.tournament_matrix
    # Triads
    is_triad = nxapi.triads.is_triad

    @staticmethod
    def convert_from_nx(graph, weight=None, *, name=None):
        import networkx as nx

        from .classes import DiGraph, Graph, MultiDiGraph, MultiGraph

        if isinstance(graph, nx.MultiDiGraph):
            return MultiDiGraph.from_networkx(graph, weight=weight)
        if isinstance(graph, nx.MultiGraph):
            return MultiGraph.from_networkx(graph, weight=weight)
        if isinstance(graph, nx.DiGraph):
            return DiGraph.from_networkx(graph, weight=weight)
        if isinstance(graph, nx.Graph):
            return Graph.from_networkx(graph, weight=weight)
        raise TypeError(f"Unsupported type of graph: {type(graph)}")

    @staticmethod
    def convert_to_nx(obj, *, name=None):
        from .classes import Graph

        if isinstance(obj, Graph):
            obj = obj.to_networkx()
        return obj

    @staticmethod
    def on_start_tests(items):
        try:
            import pytest
        except ImportError:  # pragma: no cover (import)
            return
        skip = [
            ("test_attributes", {"TestBoruvka", "test_mst.py"}),
            ("test_weight_attribute", {"TestBoruvka", "test_mst.py"}),
        ]
        for item in items:
            kset = set(item.keywords)
            for test_name, keywords in skip:
                if item.name == test_name and keywords.issubset(kset):
                    item.add_marker(
                        pytest.mark.xfail(reason="unable to handle multi-attributed graphs")
                    )



import graphblas as gb
import numpy as np
from graphblas import Matrix, Vector, binary
from graphblas.core.matrix import TransposedMatrix

################
# Classmethods #
################


def from_networkx(cls, G, weight=None, dtype=None):
    rv = cls()
    rv._key_to_id = {k: i for i, k in enumerate(G)}
    if rv._key_to_id:
        rv._A = gb.io.from_networkx(G, nodelist=rv._key_to_id, weight=weight, dtype=dtype)
    else:
        rv._A = Matrix(dtype if dtype is not None else float)
    return rv


##############
# Properties #
##############


def id_to_key(self):
    if self._id_to_key is None:
        self._id_to_key = {val: key for key, val in self._key_to_id.items()}
    return self._id_to_key


###########
# Methods #
###########


def get_property(self, name, *, mask=None):
    return self._get_property[self._cache_aliases.get(name, name)](self, mask)


def get_properties(self, names, *, mask=None):
    if isinstance(names, str):
        # Separated by commas and/or spaces
        names = [
            self._cache_aliases.get(name, name)
            for name in names.replace(" ", ",").split(",")
            if name
        ]
    else:
        names = [self._cache_aliases.get(name, name) for name in names]
    results = {
        name: self._get_property[name](self, mask)
        for name in sorted(names, key=self._property_priority.__getitem__)
    }
    return [results[name] for name in names]


def dict_to_vector(self, d, *, size=None, dtype=None, name=None):
    if d is None:
        return None
    if size is None:
        size = len(self)
    key_to_id = self._key_to_id
    indices, values = zip(*((key_to_id[key], val) for key, val in d.items()))
    return Vector.from_coo(indices, values, size=size, dtype=dtype, name=name)


def list_to_vector(self, nodes, dtype=bool, *, size=None, name=None):
    if nodes is None:
        return None
    if size is None:
        size = len(self)
    key_to_id = self._key_to_id
    index = [key_to_id[key] for key in nodes]
    return Vector.from_coo(index, True, size=size, dtype=dtype, name=name)


def list_to_mask(self, nodes, *, size=None, name="mask"):
    if nodes is None:
        return None
    return self.list_to_vector(nodes, size=size, name=name).S


def list_to_ids(self, nodes):
    if nodes is None:
        return None
    key_to_id = self._key_to_id
    return [key_to_id[key] for key in nodes]


def list_to_keys(self, indices):
    if indices is None:
        return None
    id_to_key = self.id_to_key
    return [id_to_key[idx] for idx in indices]


def set_to_vector(self, nodes, dtype=bool, *, ignore_extra=False, size=None, name=None):
    if nodes is None:
        return None
    if size is None:
        size = len(self)
    key_to_id = self._key_to_id
    if ignore_extra:
        if not isinstance(nodes, set):
            nodes = set(nodes)
        nodes = nodes & key_to_id.keys()
    index = [key_to_id[key] for key in nodes]
    return Vector.from_coo(index, True, size=size, dtype=dtype, name=name)


def vector_to_dict(self, v, *, mask=None, fillvalue=None):
    if mask is not None:
        if fillvalue is not None and v.nvals < mask.parent.nvals:
            v(mask, binary.first) << fillvalue
    elif fillvalue is not None and v.nvals < v.size:
        v(mask=~v.S) << fillvalue
    id_to_key = self.id_to_key
    return {id_to_key[index]: value for index, value in zip(*v.to_coo(sort=False))}


def vector_to_nodemap(self, v, *, mask=None, fillvalue=None):
    from .nodemap import NodeMap

    if mask is not None:
        if fillvalue is not None and v.nvals < mask.parent.nvals:
            v(mask, binary.first) << fillvalue
    elif fillvalue is not None and v.nvals < v.size:
        v(mask=~v.S) << fillvalue

    rv = NodeMap(v, key_to_id=self._key_to_id)
    rv._id_to_key = self._id_to_key
    return rv


def vector_to_nodeset(self, v):
    from .nodeset import NodeSet

    rv = NodeSet(v, key_to_id=self._key_to_id)
    rv._id_to_key = self._id_to_key
    return rv


def vector_to_set(self, v):
    id_to_key = self.id_to_key
    indices, _ = v.to_coo(values=False, sort=False)
    return {id_to_key[index] for index in indices}


def matrix_to_dicts(self, A, *, use_row_index=False, use_column_index=False):
    """Convert a Matrix to a dict of dicts of the form ``{row: {col: val}}``

    Use ``use_row_index=True`` to return the row index as keys in the dict,
    and likewise for `use_column_index=True``.

    """
    if isinstance(A, TransposedMatrix):
        # Not covered
        d = A.T.ss.export("hypercsc")
        rows = d["cols"].tolist()
        col_indices = d["row_indices"].tolist()
        use_row_index, use_column_index = use_column_index, use_row_index
    else:
        d = A.ss.export("hypercsr")
        rows = d["rows"].tolist()
        col_indices = d["col_indices"].tolist()
    indptr = d["indptr"]
    values = d["values"].tolist()
    id_to_key = self.id_to_key
    it = zip(rows, np.lib.stride_tricks.sliding_window_view(indptr, 2).tolist())
    if use_row_index and use_column_index:
        return {
            row: dict(zip(col_indices[start:stop], values[start:stop])) for row, (start, stop) in it
        }
    elif use_row_index:
        return {
            row: {
                id_to_key[col]: val for col, val in zip(col_indices[start:stop], values[start:stop])
            }
            for row, (start, stop) in it
        }
    elif use_column_index:
        return {
            id_to_key[row]: dict(zip(col_indices[start:stop], values[start:stop]))
            for row, (start, stop) in it
        }
    else:
        return {
            id_to_key[row]: {
                id_to_key[col]: val for col, val in zip(col_indices[start:stop], values[start:stop])
            }
            for row, (start, stop) in it
        }


def to_networkx(self, edge_attribute="weight"):
    import networkx as nx

    # Not covered yet, but will probably be useful soon
    if self.is_directed():
        G = nx.DiGraph()
        A = self._A
    else:
        G = nx.Graph()
        A = self.get_property("L+")
    G.add_nodes_from(self._key_to_id)
    id_to_key = self.id_to_key
    rows, cols, vals = A.to_coo()
    rows = (id_to_key[row] for row in rows.tolist())
    cols = (id_to_key[col] for col in cols.tolist())
    if edge_attribute is None:
        G.add_edges_from(zip(rows, cols))
    else:
        G.add_weighted_edges_from(zip(rows, cols, vals), weight=edge_attribute)
    # What else should we copy over?
    return G


def _cacheit(self, key, func, *args, **kwargs):
    if key not in self._cache:
        self._cache[key] = func(*args, **kwargs)
    return self._cache[key]



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph

from .exception import NetworkXError

__all__ = ["descendants", "ancestors"]


def descendants(G, source):
    G = to_graph(G)
    try:
        result = algorithms.descendants(G, source)
    except KeyError as e:
        raise NetworkXError(*e.args) from e
    else:
        return G.vector_to_nodeset(result)


def ancestors(G, source):
    G = to_graph(G)
    try:
        result = algorithms.ancestors(G, source)
    except KeyError as e:
        raise NetworkXError(*e.args) from e
    else:
        return G.vector_to_nodeset(result)



import numpy as np
from graphblas import select

__all__ = ["is_tournament", "score_sequence", "tournament_matrix"]


def is_tournament(G):
    A = G._A
    if A.nvals != A.nrows * (A.ncols - 1) // 2 or G.get_property("has_self_edges"):
        return False
    # Alternative: do `select.triu(A.T).new(mask=A.S)` and don't check "has_self_edges"
    val = select.triu(A.T, 1).new(mask=A.S)
    return val.nvals == 0


def score_sequence(G):
    degrees = G.get_property("row_degrees+")
    _, values = degrees.to_coo(indices=False, sort=False)
    values.sort()
    if degrees.nvals != degrees.size:
        values = np.pad(values, (degrees.size - degrees.nvals, 0))
    return values


def tournament_matrix(G):
    A = G._A
    return (A - A.T).new(name="tournament_matrix")



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph

__all__ = ["is_isolate", "isolates", "number_of_isolates"]


def is_isolate(G, n):
    G = to_graph(G)
    return algorithms.is_isolate(G, n)


def isolates(G):
    G = to_graph(G)
    result = algorithms.isolates(G)
    return G.vector_to_nodeset(result)  # Return type is iterable


def number_of_isolates(G):
    G = to_graph(G)
    return algorithms.number_of_isolates(G)



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph

__all__ = [
    "cut_size",
    "volume",
    "normalized_cut_size",
    "conductance",
    "edge_expansion",
    "mixing_expansion",
    "node_expansion",
    "boundary_expansion",
]


def cut_size(G, S, T=None, weight=None):
    is_multigraph = G.is_multigraph()
    G = to_graph(G, weight=weight)
    S = G.set_to_vector(S, ignore_extra=True)
    T = G.set_to_vector(T, ignore_extra=True)
    return algorithms.cut_size(G, S, T, is_weighted=is_multigraph or weight is not None)


def volume(G, S, weight=None):
    is_multigraph = G.is_multigraph()
    G = to_graph(G, weight=weight)
    S = G.list_to_vector(S)
    return algorithms.volume(G, S, weighted=is_multigraph or weight is not None)


def normalized_cut_size(G, S, T=None, weight=None):
    G = to_graph(G, weight=weight)
    S = G.set_to_vector(S, ignore_extra=True)
    if T is None:
        T = (~S.S).new()
    else:
        T = G.set_to_vector(T, ignore_extra=True)
    return algorithms.normalized_cut_size(G, S, T)


def conductance(G, S, T=None, weight=None):
    G = to_graph(G, weight=weight)
    S = G.set_to_vector(S, ignore_extra=True)
    if T is None:
        T = (~S.S).new()
    else:
        T = G.set_to_vector(T, ignore_extra=True)
    return algorithms.conductance(G, S, T)


def edge_expansion(G, S, T=None, weight=None):
    G = to_graph(G, weight=weight)
    S = G.set_to_vector(S, ignore_extra=True)
    T = G.set_to_vector(T, ignore_extra=True)
    return algorithms.edge_expansion(G, S, T)


def mixing_expansion(G, S, T=None, weight=None):
    G = to_graph(G, weight=weight)
    S = G.set_to_vector(S, ignore_extra=True)
    T = G.set_to_vector(T, ignore_extra=True)
    return algorithms.mixing_expansion(G, S, T)


def node_expansion(G, S):
    G = to_graph(G)
    S = G.list_to_vector(S)
    return algorithms.node_expansion(G, S)


def boundary_expansion(G, S):
    G = to_graph(G)
    S = G.set_to_vector(S, ignore_extra=True)
    return algorithms.boundary_expansion(G, S)



from graphblas import binary

from .exceptions import EmptyGraphError

__all__ = ["reciprocity", "overall_reciprocity"]


def reciprocity(G, mask=None):
    overlap, total_degrees = G.get_properties("recip_degrees+ total_degrees+", mask=mask)
    return binary.truediv(2 * overlap | total_degrees, left_default=0, right_default=0).new(
        name="reciprocity"
    )


def overall_reciprocity(G):
    n_all_edge = G._A.nvals
    if n_all_edge == 0:
        raise EmptyGraphError("Not defined for empty graphs")
    n_overlap_edges, has_self_edges = G.get_properties("total_recip+ has_self_edges")
    if has_self_edges:
        n_overlap_edges -= G.get_property("diag").nvals
    return n_overlap_edges / n_all_edge



from graphblas_algorithms.classes.digraph import DiGraph
from graphblas_algorithms.classes.graph import Graph

__all__ = ["is_triad"]


def is_triad(G):
    return (
        isinstance(G, (Graph, DiGraph))
        and G.is_directed()
        and G.order() == 3
        and not G.get_property("has_self_edges")
    )



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph
from graphblas_algorithms.utils import not_implemented_for

from ..exception import NetworkXError, NetworkXPointlessConcept, PowerIterationFailedConvergence

__all__ = ["eigenvector_centrality"]


@not_implemented_for("multigraph")
def eigenvector_centrality(G, max_iter=100, tol=1.0e-6, nstart=None, weight=None):
    G = to_graph(G, weight=weight, dtype=float)
    if len(G) == 0:
        raise NetworkXPointlessConcept("cannot compute centrality for the null graph")
    x = G.dict_to_vector(nstart, dtype=float, name="nstart")
    try:
        result = algorithms.eigenvector_centrality(G, max_iter=max_iter, tol=tol, nstart=x)
    except algorithms.exceptions.ConvergenceFailure as e:
        raise PowerIterationFailedConvergence(*e.args) from e
    except algorithms.exceptions.GraphBlasAlgorithmException as e:
        raise NetworkXError(*e.args) from e
    return G.vector_to_nodemap(result)



from graphblas import Vector

from graphblas_algorithms.algorithms._helpers import is_converged, normalize
from graphblas_algorithms.algorithms.exceptions import ConvergenceFailure

__all__ = ["hits"]


def hits(G, max_iter=100, tol=1.0e-8, nstart=None, normalized=True, *, with_authority=False):
    """HITS algorithms with additional parameter `with_authority`.

    When `with_authority` is True, the authority matrix, ``A.T @ A`` will be
    created and used. This may be faster, but requires more memory.
    """
    N = len(G)
    h = Vector(float, N, name="h")
    a = Vector(float, N, name="a")
    if N == 0:
        return h, a
    if nstart is None:
        h << 1.0 / N
    else:
        h << nstart
        denom = h.reduce().get(0)
        h *= 1.0 / denom

    # Power iteration: make up to max_iter iterations
    A = G._A
    if with_authority:
        a, h = h, a
        ATA = (A.T @ A).new(name="ATA")  # Authority matrix
        aprev = Vector(float, N, name="a_prev")
        for _ in range(max_iter):
            aprev, a = a, aprev
            a << ATA @ aprev
            normalize(a, "Linf")
            if is_converged(aprev, a, tol):
                h << A @ a
                break
        else:
            raise ConvergenceFailure(max_iter)
    else:
        hprev = Vector(float, N, name="h_prev")
        for _ in range(max_iter):
            hprev, h = h, hprev
            a << hprev @ A
            h << A @ a
            normalize(h, "Linf")
            if is_converged(hprev, h, tol):
                break
        else:
            raise ConvergenceFailure(max_iter)
    if normalized:
        normalize(h, "L1")
        normalize(a, "L1")
    elif with_authority:
        normalize(h, "Linf")
    else:
        normalize(a, "Linf")
    h.name = "hits_h"
    a.name = "hits_a"
    return h, a



from graphblas import binary
from graphblas.semiring import any_pair

__all__ = ["edge_boundary", "node_boundary"]


def edge_boundary(G, nbunch1, nbunch2=None, *, is_weighted=False):
    if is_weighted:
        B = binary.second(nbunch1 & G._A).new(name="boundary")
    else:
        B = binary.pair(nbunch1 & G._A).new(name="boundary")
    if nbunch2 is None:
        # Default nbunch2 is the complement of nbunch1.
        # We get the row_degrees to better handle hypersparse data.
        nbunch2 = G.get_property("row_degrees+", mask=~nbunch1.S)
    if is_weighted:
        B << binary.first(B & nbunch2)
    else:
        B << binary.pair(B & nbunch2)
    return B


def node_boundary(G, nbunch1, *, mask=None):
    if mask is None:
        mask = ~nbunch1.S
    else:
        mask = mask & (~nbunch1.S)
    bdy = any_pair(nbunch1 @ G._A).new(mask=mask, name="boundary")
    return bdy



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_directed_graph, to_graph
from graphblas_algorithms.utils import not_implemented_for

__all__ = ["degree_centrality", "in_degree_centrality", "out_degree_centrality"]


def degree_centrality(G):
    G = to_graph(G)
    result = algorithms.degree_centrality(G)
    return G.vector_to_nodemap(result, fillvalue=0.0)


@not_implemented_for("undirected")
def in_degree_centrality(G):
    G = to_directed_graph(G)
    result = algorithms.in_degree_centrality(G)
    return G.vector_to_nodemap(result, fillvalue=0.0)


@not_implemented_for("undirected")
def out_degree_centrality(G):
    G = to_directed_graph(G)
    result = algorithms.out_degree_centrality(G)
    return G.vector_to_nodemap(result, fillvalue=0.0)



from graphblas import Matrix, binary

__all__ = ["is_simple_path"]


def is_simple_path(G, nodes):
    if len(nodes) == 0:
        return False
    if len(nodes) == 1:
        return nodes[0] in G
    A = G._A
    if A.nvals < len(nodes) - 1:
        return False
    key_to_id = G._key_to_id
    indices = [key_to_id[key] for key in nodes if key in key_to_id]
    if len(indices) != len(nodes) or len(indices) > len(set(indices)):
        return False
    # Check all steps in path at once
    P = Matrix.from_coo(indices[:-1], indices[1:], True, nrows=A.nrows, ncols=A.ncols)
    P << binary.second(A & P)
    return P.nvals == len(indices) - 1
    # Alternative
    # it = iter(indices)
    # prev = next(it)
    # for cur in it:
    #     if (prev, cur) not in A:
    #         return False
    #     prev = cur
    # return True



from collections.abc import MutableMapping

from graphblas import Vector, monoid

from . import _utils


class NodeMap(MutableMapping):
    def __init__(self, v, *, key_to_id=None):
        self.vector = v
        if key_to_id is None:
            self._key_to_id = {i: i for i in range(v.size)}
        else:
            self._key_to_id = key_to_id
        self._id_to_key = None

    id_to_key = property(_utils.id_to_key)
    # get_property = _utils.get_property
    # get_properties = _utils.get_properties
    dict_to_vector = _utils.dict_to_vector
    list_to_vector = _utils.list_to_vector
    list_to_mask = _utils.list_to_mask
    list_to_ids = _utils.list_to_ids
    list_to_keys = _utils.list_to_keys
    matrix_to_dicts = _utils.matrix_to_dicts
    set_to_vector = _utils.set_to_vector
    # to_networkx = _utils.to_networkx
    vector_to_dict = _utils.vector_to_dict
    vector_to_nodemap = _utils.vector_to_nodemap
    vector_to_nodeset = _utils.vector_to_nodeset
    vector_to_set = _utils.vector_to_set
    # _cacheit = _utils._cacheit

    # Requirements for MutableMapping
    def __delitem__(self, key):
        idx = self._key_to_id[key]
        del self.vector[idx]

    def __getitem__(self, key):
        idx = self._key_to_id[key]
        if (rv := self.vector.get(idx)) is not None:
            return rv
        raise KeyError(key)

    def __iter__(self):
        # Slow if we iterate over one; fast if we iterate over all
        return map(
            self.id_to_key.__getitem__, self.vector.to_coo(values=False, sort=False)[0].tolist()
        )

    def __len__(self):
        return self.vector.nvals

    def __setitem__(self, key, val):
        idx = self._key_to_id[key]
        self.vector[idx] = val

    # Override other MutableMapping methods
    def __contains__(self, key):
        idx = self._key_to_id[key]
        return idx in self.vector

    def __eq__(self, other):
        if isinstance(other, NodeMap):
            return self.vector.isequal(other.vector) and self._key_to_id == other._key_to_id
        return super().__eq__(other)

    def clear(self):
        self.vector.clear()

    def get(self, key, default=None):
        idx = self._key_to_id[key]
        return self.vector.get(idx, default)

    # items
    # keys
    # pop

    def popitem(self):
        v = self.vector
        try:
            idx, value = next(v.ss.iteritems())
        except StopIteration:
            raise KeyError from None
        del v[idx]
        return self.id_to_key[idx], value

    def setdefault(self, key, default=None):
        idx = self._key_to_id[key]
        if (value := self.vector.get(idx)) is not None:
            return value
        self.vector[idx] = default
        return default

    # update
    # values


class VectorMap(MutableMapping):
    def __init__(self, v):
        self.vector = v

    # Requirements for MutableMapping
    def __delitem__(self, key):
        del self.vector[key]

    def __getitem__(self, key):
        return self.vector.get(key)

    def __iter__(self):
        # Slow if we iterate over one; fast if we iterate over all
        return iter(self.vector.to_coo(values=False, sort=False)[0].tolist())

    def __len__(self):
        return self.vector.nvals

    def __setitem__(self, key, val):
        self.vector[key] = val

    # Override other MutableMapping methods
    def __contains__(self, key):
        return key in self.vector

    def __eq__(self, other):
        if isinstance(other, VectorMap):
            return self.vector.isequal(other.vector)
        if isinstance(other, Vector):
            return self.vector.isequal(other)
        if isinstance(other, NodeMap):
            return self.vector.isequal(other.vector) and other._key_to_id == {
                i: i for i in range(self.vector.size)
            }
        return super().__eq__(other)

    def clear(self):
        self.vector.clear()

    def get(self, key, default=None):
        return self.vector.get(key, default)

    # items
    # keys
    # pop

    def popitem(self):
        v = self.vector
        try:
            idx, value = next(v.ss.iteritems())
        except StopIteration:
            raise KeyError from None
        del v[idx]
        return idx, value

    def setdefault(self, key, default=None):
        if (value := self.vector.get(key)) is not None:
            return value
        self.vector[key] = default
        return default

    # update
    # values


class VectorNodeMap(MutableMapping):
    def __init__(self, A, *, key_to_id=None):
        self.matrix = A
        if key_to_id is None:
            self._key_to_id = {i: i for i in range(A.size)}
        else:
            self._key_to_id = key_to_id
        self._id_to_key = None
        self._rows = None

    def _get_rows(self):
        if self._rows is None:
            self._rows = self.matrix.reduce_rowwise(monoid.any).new()
            self._rows(self._rows.S) << 1  # Make iso-valued
        return self._rows

    id_to_key = property(_utils.id_to_key)
    # get_property = _utils.get_property
    # get_properties = _utils.get_properties
    dict_to_vector = _utils.dict_to_vector
    list_to_vector = _utils.list_to_vector
    list_to_mask = _utils.list_to_mask
    list_to_ids = _utils.list_to_ids
    list_to_keys = _utils.list_to_keys
    matrix_to_dicts = _utils.matrix_to_dicts
    set_to_vector = _utils.set_to_vector
    # to_networkx = _utils.to_networkx
    vector_to_dict = _utils.vector_to_dict
    vector_to_nodemap = _utils.vector_to_nodemap
    vector_to_nodeset = _utils.vector_to_nodeset
    vector_to_set = _utils.vector_to_set
    # _cacheit = _utils._cacheit

    # Requirements for MutableMapping
    def __delitem__(self, key):
        idx = self._key_to_id[key]
        del self.matrix[idx, :]
        if self._rows is not None:
            del self._rows[idx]

    def __getitem__(self, key):
        idx = self._key_to_id[key]
        if self._get_rows().get(idx) is None:
            raise KeyError(key)
        return VectorMap(self.matrix[idx, :].new())

    def __iter__(self):
        # Slow if we iterate over one; fast if we iterate over all
        return map(
            self.id_to_key.__getitem__,
            self._get_rows().to_coo(values=False, sort=False)[0].tolist(),
        )

    def __len__(self):
        return self._get_rows().nvals

    def __setitem__(self, key, val):
        idx = self._key_to_id[key]
        if isinstance(val, VectorMap):
            val = val.vector
        elif isinstance(val, dict):
            val = Vector.from_dict(val, self.matrix.dtype, size=self.matrix.ncols)
        else:
            raise TypeError()
        if val.nvals == 0:
            del self.matrix[idx, :]
            if self._rows is not None:
                del self._rows[idx]
        else:
            self.matrix[idx, :] = val
            if self._rows is not None:
                self._rows[idx] = 1

    # Override other MutableMapping methods
    def __contains__(self, key):
        idx = self._key_to_id[key]
        return idx in self._get_rows()

    def __eq__(self, other):
        if isinstance(other, VectorNodeMap):
            return self.matrix.isequal(other.matrix) and self._key_to_id == other._key_to_id
        return super().__eq__(other)

    def clear(self):
        self.matrix.clear()
        self._rows = None

    def get(self, key, default=None):
        idx = self._key_to_id[key]
        if self._get_rows().get(idx) is None:
            return default
        return VectorMap(self.matrix[idx, :].new())

    # items
    # keys
    # pop

    def popitem(self):
        rows = self._get_rows()
        try:
            idx = next(rows.ss.iterkeys())
        except StopIteration:
            raise KeyError from None
        value = VectorMap(self.matrix[idx, :].new())
        del self.matrix[idx, :]
        del rows[idx]
        return self.id_to_key[idx], value

    # setdefault
    # update
    # values



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_directed_graph
from graphblas_algorithms.utils import not_implemented_for

from .exception import NetworkXError

__all__ = ["reciprocity", "overall_reciprocity"]


@not_implemented_for("undirected", "multigraph")
def reciprocity(G, nodes=None):
    if nodes is None:
        return overall_reciprocity(G)
    G = to_directed_graph(G, dtype=bool)
    if nodes in G:
        mask = G.list_to_mask([nodes])
        result = algorithms.reciprocity(G, mask=mask)
        rv = result.get(G._key_to_id[nodes])
        if rv is None:
            raise NetworkXError("Not defined for isolated nodes.")
        else:
            return rv
    else:
        mask = G.list_to_mask(nodes)
        result = algorithms.reciprocity(G, mask=mask)
        return G.vector_to_nodemap(result, mask=mask)


@not_implemented_for("undirected", "multigraph")
def overall_reciprocity(G):
    G = to_directed_graph(G, dtype=bool)
    try:
        return algorithms.overall_reciprocity(G)
    except algorithms.exceptions.EmptyGraphError as e:
        raise NetworkXError("Not defined for empty graphs") from e



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.graph import to_undirected_graph
from graphblas_algorithms.utils import not_implemented_for

__all__ = ["k_truss"]


@not_implemented_for("directed")
@not_implemented_for("multigraph")
def k_truss(G, k):
    G = to_undirected_graph(G, dtype=bool)
    result = algorithms.k_truss(G, k)
    return result



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph

__all__ = ["is_dominating_set"]


def is_dominating_set(G, nbunch):
    G = to_graph(G)
    v = G.set_to_vector(nbunch, ignore_extra=True)
    return algorithms.is_dominating_set(G, v)



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph

__all__ = []


def mutual_weight(G, u, v, weight=None):
    G = to_graph(G, weight=weight)
    return algorithms.mutual_weight(G, u, v)



""" Test that `graphblas.nxapi` structure matches that of networkx.

This currently checks the locations and names of all networkx-dispatchable functions
that are implemented by `graphblas_algorithms`.  It ignores names that begin with `_`.

The `test_dispatched_funcs_in_nxap` test below will say what to add and delete under `nxapi`.

We should consider removing any test here that becomes too much of a nuisance.
For now, though, let's try to match and stay up-to-date with NetworkX!

"""
import sys
from collections import namedtuple

import pytest

import graphblas_algorithms as ga

try:
    import networkx as nx
except ImportError:
    pytest.skip(
        "Matching networkx namespace requires networkx to be installed", allow_module_level=True
    )
else:
    from networkx.classes import backends


def isdispatched(func):
    """Can this NetworkX function dispatch to other backends?"""
    return (
        callable(func) and hasattr(func, "dispatchname") and func.__module__.startswith("networkx")
    )


def dispatchname(func):
    """The dispatched name of the dispatchable NetworkX function"""
    # Haha, there should be a better way to get this
    if not isdispatched(func):
        raise ValueError(f"Function is not dispatched in NetworkX: {func.__name__}")
    return func.dispatchname


def fullname(func):
    return f"{func.__module__}.{func.__name__}"


NameInfo = namedtuple("NameInfo", ["dispatchname", "fullname", "curpath"])


@pytest.fixture(scope="module")
def nx_info():
    rv = {}  # {modulepath: {dispatchname: NameInfo}}
    for modname, module in sys.modules.items():
        cur = {}
        if not modname.startswith("networkx.") and modname != "networkx" or "tests" in modname:
            continue
        for key, val in vars(module).items():
            if not key.startswith("_") and isdispatched(val):
                dname = dispatchname(val)
                cur[dname] = NameInfo(dname, fullname(val), f"{modname}.{key}")
        if cur:
            rv[modname] = cur
    return rv


@pytest.fixture(scope="module")
def gb_info():
    rv = {}  # {modulepath: {dispatchname: NameInfo}}
    from graphblas_algorithms import nxapi
    from graphblas_algorithms.interface import Dispatcher

    ga_map = {
        fullname(val): key
        for key, val in vars(Dispatcher).items()
        if callable(val) and fullname(val).startswith("graphblas_algorithms.nxapi.")
    }
    for modname, module in sys.modules.items():
        cur = {}
        if not modname.startswith("graphblas_algorithms.nxapi") or "tests" in modname:
            continue
        for key, val in vars(module).items():
            try:
                fname = fullname(val)
            except Exception:
                continue
            if key.startswith("_") or fname not in ga_map:
                continue
            dname = ga_map[fname]
            cur[dname] = NameInfo(dname, fname, f"{modname}.{key}")
        if cur:
            rv[modname] = cur
    return rv


@pytest.fixture(scope="module")
def nx_names_to_info(nx_info):
    rv = {}  # {dispatchname: {NameInfo}}
    for names in nx_info.values():
        for name, info in names.items():
            if name not in rv:
                rv[name] = set()
            rv[name].add(info)
    return rv


@pytest.fixture(scope="module")
def gb_names_to_info(gb_info):
    rv = {}  # {dispatchname: {NameInfo}}
    for names in gb_info.values():
        for name, info in names.items():
            if name not in rv:
                rv[name] = set()
            rv[name].add(info)
    return rv


@pytest.mark.checkstructure
def test_nonempty(nx_info, gb_info, nx_names_to_info, gb_names_to_info):
    assert len(nx_info) > 15
    assert len(gb_info) > 15
    assert len(nx_names_to_info) > 30
    assert len(gb_names_to_info) > 30


def nx_to_gb_info(info):
    gb = "graphblas_algorithms.nxapi"
    return NameInfo(
        info[0],
        info[1].replace("networkx.algorithms", gb).replace("networkx", gb),
        info[2].replace("networkx.algorithms", gb).replace("networkx", gb),
    )


@pytest.mark.checkstructure
def test_dispatched_funcs_in_nxapi(nx_names_to_info, gb_names_to_info):
    """Are graphblas_algorithms functions in the correct locations in nxapi?"""
    failing = False
    for name in nx_names_to_info.keys() & gb_names_to_info.keys():
        nx_paths = {nx_to_gb_info(info) for info in nx_names_to_info[name]}
        gb_paths = gb_names_to_info[name]
        if nx_paths != gb_paths:  # pragma: no cover
            failing = True
            if missing := (nx_paths - gb_paths):
                from_ = ":".join(next(iter(missing))[1].rsplit(".", 1))
                print(f"Add `{name}` from `{from_}` here:")
                for _, _, path in sorted(missing):
                    print("   ", ":".join(path.rsplit(".", 1)))
            if extra := (gb_paths - nx_paths):
                from_ = ":".join(next(iter(extra))[1].rsplit(".", 1))
                print(f"Remove `{name}` from `{from_}` here:")
                for _, _, path in sorted(extra):
                    print("   ", ":".join(path.rsplit(".", 1)))
    if failing:  # pragma: no cover
        raise AssertionError()



from graphblas import Matrix

__all__ = ["not_implemented_for"]


def not_implemented_for(*graph_types):
    import networkx.utils.decorators

    rv = networkx.utils.decorators.not_implemented_for(*graph_types)
    func = rv._func

    def inner(g):
        if not isinstance(g, Matrix):
            return func(g)
        # Let Matrix objects pass through and check later.
        # We could check now and convert to appropriate graph type.
        return g

    rv._func = inner
    return rv



from graphblas import op
from graphblas.core import operator


def get_reduce_to_vector(key, opname, methodname):
    op_ = op.from_string(opname)
    op_, opclass = operator.find_opclass(op_)
    keybase = key[:-1]
    if key[-1] == "-":

        def get_reduction(G, mask=None):
            cache = G._cache
            if mask is not None:
                if key in cache:
                    return cache[key].dup(mask=mask)
                elif cache.get("has_self_edges") is False and f"{keybase}+" in cache:
                    cache[key] = cache[f"{keybase}+"]
                    return cache[key].dup(mask=mask)
                elif "offdiag" in cache:
                    return getattr(cache["offdiag"], methodname)(op_).new(mask=mask, name=key)
                elif (
                    "L-" in cache
                    and "U-" in cache
                    and opclass in {"BinaryOp", "Monoid"}
                    and G.get_property("has_self_edges")
                ):
                    return op_(
                        getattr(cache["L-"], methodname)(op_).new(mask=mask)
                        | getattr(cache["U-"], methodname)(op_).new(mask=mask)
                    ).new(name=key)
                elif not G.get_property("has_self_edges"):
                    return G.get_property(f"{keybase}+", mask=mask)
                else:
                    return getattr(G.get_property("offdiag"), methodname)(op_).new(
                        mask=mask, name=key
                    )
            if key not in cache:
                if cache.get("has_self_edges") is False and f"{keybase}+" in cache:
                    cache[key] = cache[f"{keybase}+"]
                elif "offdiag" in cache:
                    cache[key] = getattr(cache["offdiag"], methodname)(op_).new(name=key)
                elif (
                    "L-" in cache
                    and "U-" in cache
                    and opclass in {"BinaryOp", "Monoid"}
                    and G.get_property("has_self_edges")
                ):
                    cache[key] = op_(
                        getattr(cache["L-"], methodname)(op_)
                        | getattr(cache["U-"], methodname)(op_)
                    ).new(name=key)
                elif not G.get_property("has_self_edges"):
                    cache[key] = G.get_property(f"{keybase}+")
                else:
                    cache[key] = getattr(G.get_property("offdiag"), methodname)(op_).new(name=key)
            if (
                "has_self_edges" not in cache
                and f"{keybase}+" in cache
                and cache[key].nvals != cache[f"{keybase}+"].nvals
            ):
                cache["has_self_edges"] = True
            elif cache.get("has_self_edges") is False:
                cache[f"{keybase}+"] = cache[key]
            return cache[key]

    else:

        def get_reduction(G, mask=None):
            A = G._A
            cache = G._cache
            if mask is not None:
                if key in cache:
                    return cache[key].dup(mask=mask)
                elif cache.get("has_self_edges") is False and f"{keybase}-" in cache:
                    cache[key] = cache[f"{keybase}-"]
                    return cache[key].dup(mask=mask)
                elif methodname == "reduce_columnwise" and "AT" in cache:
                    return cache["AT"].reduce_rowwise(op_).new(mask=mask, name=key)
                else:
                    return getattr(A, methodname)(op_).new(mask=mask, name=key)
            if key not in cache:
                if cache.get("has_self_edges") is False and f"{keybase}-" in cache:
                    cache[key] = cache[f"{keybase}-"]
                elif methodname == "reduce_columnwise" and "AT" in cache:
                    cache[key] = cache["AT"].reduce_rowwise(op_).new(name=key)
                else:
                    cache[key] = getattr(A, methodname)(op_).new(name=key)
            if (
                "has_self_edges" not in cache
                and f"{keybase}-" in cache
                and cache[key].nvals != cache[f"{keybase}-"].nvals
            ):
                cache["has_self_edges"] = True
            elif cache.get("has_self_edges") is False:
                cache[f"{keybase}-"] = cache[key]
            return cache[key]

    return get_reduction


def get_reduce_to_scalar(key, opname):
    op_ = op.from_string(opname)
    op_, opclass = operator.find_opclass(op_)
    keybase = key[:-1]
    if key[-1] == "-":

        def get_reduction(G, mask=None):
            cache = G._cache
            if key not in cache:
                if cache.get("has_self_edges") is False and f"{keybase}+" in cache:
                    cache[key] = cache[f"{keybase}+"]
                elif f"{opname}_rowwise-" in cache:
                    cache[key] = cache[f"{opname}_rowwise-"].reduce(op_).new(name=key)
                elif f"{opname}_columnwise-" in cache:
                    cache[key] = cache[f"{opname}_columnwise-"].reduce(op_).new(name=key)
                elif cache.get("has_self_edges") is False and f"{opname}_rowwise+" in cache:
                    cache[key] = cache[f"{opname}_rowwise+"].reduce(op_).new(name=key)
                elif cache.get("has_self_edges") is False and f"{opname}_columnwise+" in cache:
                    cache[key] = cache[f"{opname}_columnwise+"].reduce(op_).new(name=key)
                elif "offdiag" in cache:
                    cache[key] = cache["offdiag"].reduce_scalar(op_).new(name=key)
                elif (
                    "L-" in cache
                    and "U-" in cache
                    and opclass in {"BinaryOp", "Monoid"}
                    and G.get_property("has_self_edges")
                ):
                    return op_(
                        cache["L-"].reduce(op_)._as_vector() | cache["U-"].reduce(op_)._as_vector()
                    )[0].new(name=key)
                elif not G.get_property("has_self_edges"):
                    cache[key] = G.get_property(f"{keybase}+")
                else:
                    cache[key] = G.get_property("offdiag").reduce_scalar(op_).new(name=key)
            if (
                "has_self_edges" not in cache
                and f"{keybase}+" in cache
                and cache[key] != cache[f"{keybase}+"]
            ):
                cache["has_self_edges"] = True
            elif cache.get("has_self_edges") is False:
                cache[f"{keybase}+"] = cache[key]
            return cache[key]

    else:

        def get_reduction(G, mask=None):
            A = G._A
            cache = G._cache
            if key not in cache:
                if cache.get("has_self_edges") is False and f"{keybase}-" in cache:
                    cache[key] = cache[f"{keybase}-"]
                elif f"{opname}_rowwise+" in cache:
                    cache[key] = cache[f"{opname}_rowwise+"].reduce(op_).new(name=key)
                elif f"{opname}_columnwise+" in cache:
                    cache[key] = cache[f"{opname}_columnwise+"].reduce(op_).new(name=key)
                elif cache.get("has_self_edges") is False and f"{opname}_rowwise-" in cache:
                    cache[key] = cache[f"{opname}_rowwise-"].reduce(op_).new(name=key)
                elif cache.get("has_self_edges") is False and f"{opname}_columnwise-" in cache:
                    cache[key] = cache[f"{opname}_columnwise-"].reduce(op_).new(name=key)
                else:
                    cache[key] = A.reduce_scalar(op_).new(name=key)
            if (
                "has_self_edges" not in cache
                and f"{keybase}-" in cache
                and cache[key] != cache[f"{keybase}-"]
            ):
                cache["has_self_edges"] = True
            elif cache.get("has_self_edges") is False:
                cache[f"{keybase}-"] = cache[key]
            return cache[key]

    return get_reduction



from graphblas import Scalar, Vector
from graphblas.core.utils import output_type
from graphblas.semiring import plus_first, plus_times

from graphblas_algorithms.algorithms._helpers import is_converged, normalize
from graphblas_algorithms.algorithms.exceptions import (
    ConvergenceFailure,
    GraphBlasAlgorithmException,
)

__all__ = ["katz_centrality"]


def katz_centrality(
    G,
    alpha=0.1,
    beta=1.0,
    max_iter=1000,
    tol=1.0e-6,
    nstart=None,
    normalized=True,
    name="katz_centrality",
):
    N = len(G)
    x = Vector(float, N, name="x")
    if nstart is None:
        x << 0.0
    else:
        x << nstart
    if output_type(beta) is not Vector:
        b = Scalar.from_value(beta, dtype=float, name="beta")
    else:
        b = beta
        if b.nvals != N:
            raise GraphBlasAlgorithmException("beta must have a value for every node")

    A = G._A
    if A.ss.is_iso:
        # Fold iso-value into alpha
        alpha *= A.ss.iso_value.get(1.0)
        semiring = plus_first[float]
    else:
        semiring = plus_times[float]

    # Power iteration: make up to max_iter iterations
    xprev = Vector(float, N, name="x_prev")
    for _ in range(max_iter):
        xprev, x = x, xprev
        # x << alpha * semiring(xprev @ A) + beta
        x << semiring(xprev @ A)
        x *= alpha
        x += b
        if is_converged(xprev, x, tol):  # sum(abs(xprev - x)) < N * tol
            x.name = name
            if normalized:
                normalize(x, "L2")
            return x
    raise ConvergenceFailure(max_iter)



from collections.abc import MutableSet

from graphblas.semiring import lor_pair, plus_pair

from . import _utils


class NodeSet(MutableSet):
    def __init__(self, v, *, key_to_id=None):
        self.vector = v
        if key_to_id is None:
            self._key_to_id = {i: i for i in range(v.size)}
        else:
            self._key_to_id = key_to_id
        self._id_to_key = None

    id_to_key = property(_utils.id_to_key)
    # get_property = _utils.get_property
    # get_properties = _utils.get_properties
    dict_to_vector = _utils.dict_to_vector
    list_to_vector = _utils.list_to_vector
    list_to_mask = _utils.list_to_mask
    list_to_ids = _utils.list_to_ids
    list_to_keys = _utils.list_to_keys
    matrix_to_dicts = _utils.matrix_to_dicts
    set_to_vector = _utils.set_to_vector
    # to_networkx = _utils.to_networkx
    vector_to_dict = _utils.vector_to_dict
    vector_to_nodemap = _utils.vector_to_nodemap
    vector_to_nodeset = _utils.vector_to_nodeset
    vector_to_set = _utils.vector_to_set
    # _cacheit = _utils._cacheit

    # Requirements for MutableSet
    def __contains__(self, x):
        idx = self._key_to_id[x]
        return idx in self.vector

    def __iter__(self):
        # Slow if we iterate over one; fast if we iterate over all
        return map(
            self.id_to_key.__getitem__, self.vector.to_coo(values=False, sort=False)[0].tolist()
        )

    def __len__(self):
        return self.vector.nvals

    def add(self, value):
        idx = self._key_to_id[value]
        self.vector[idx] = True

    def discard(self, value):
        idx = self._key_to_id[value]
        del self.vector[idx]

    # Override other MutableSet methods
    def __eq__(self, other):
        if isinstance(other, NodeSet):
            a = self.vector
            b = other.vector
            return (
                a.size == b.size
                and (nvals := a.nvals) == b.nvals
                and plus_pair(a @ b).get(0) == nvals
                and self._key_to_id == other._key_to_id
            )
        return super().__eq__(other)

    # __and__
    # __or__
    # __sub__
    # __xor__

    def clear(self):
        self.vector.clear()

    def isdisjoin(self, other):
        if isinstance(other, NodeSet):
            return not lor_pair(self.vector @ other.vector)
        return super().isdisjoint(other)

    def pop(self):
        try:
            idx = next(self.vector.ss.iterkeys())
        except StopIteration:
            raise KeyError from None
        del self.vector[idx]
        return self.id_to_key[idx]

    def remove(self, value):
        idx = self._key_to_id[value]
        if idx not in self.vector:
            raise KeyError(value)
        del self.vector[idx]

    def _from_iterable(self, it):
        # The elements in the iterable must be contained within key_to_id
        rv = object.__new__(type(self))
        rv._key_to_id = self._key_to_id
        rv._id_to_key = self._id_to_key
        rv.vector = rv.set_to_vector(it, size=self.vector.size)
        return rv

    # Add more set methods (as needed)
    def union(self, *args):
        return set(self).union(*args)  # TODO: can we make this better?



import numpy as np
from graphblas import Matrix, Vector, binary, replace, unary
from graphblas.semiring import plus_first, plus_pair, plus_times

__all__ = [
    "single_triangle",
    "triangles",
    "total_triangles",
    "transitivity",
    "transitivity_directed",
    "clustering",
    "clustering_directed",
    "single_clustering",
    "single_clustering_directed",
    "average_clustering",
    "average_clustering_directed",
    "single_square_clustering",
    "square_clustering",
    "generalized_degree",
    "single_generalized_degree",
]


def single_triangle(G, node, *, weighted=False):
    index = G._key_to_id[node]
    r = G._A[index, :].new()
    # Pretty much all the time is spent here taking TRIL, which is used to ignore self-edges
    L = G.get_property("L-")
    if G.get_property("has_self_edges"):
        del r[index]  # Ignore self-edges
    if weighted:
        maxval = G.get_property("max_element-")
        L = unary.cbrt(L / maxval)
        r = unary.cbrt(r / maxval)
        semiring = plus_times
    else:
        semiring = plus_pair
    val = semiring(L @ r).new(mask=r.S)
    if weighted:
        val *= r
    return val.reduce().get(0)


def triangles(G, *, weighted=False, mask=None):
    # Ignores self-edges
    # Can we apply the mask earlier in the computation?
    L, U = G.get_properties("L- U-")
    if weighted:
        maxval = G.get_property("max_element-")
        L = unary.cbrt(L / maxval)
        U = unary.cbrt(U / maxval)
        semiring = plus_times
    else:
        semiring = plus_pair
    C = semiring(L @ L.T).new(mask=L.S)
    D = semiring(U @ L.T).new(mask=U.S)
    if weighted:
        C *= L
        D *= U
    return (
        C.reduce_rowwise().new(mask=mask)
        + C.reduce_columnwise().new(mask=mask)
        + D.reduce_rowwise().new(mask=mask)
    ).new(name="triangles")


def total_triangles(G):
    # We use SandiaDot method, because it's usually the fastest on large graphs.
    # For smaller graphs, Sandia method is usually faster: plus_pair(L @ L).new(mask=L.S)
    L, U = G.get_properties("L- U-")
    return plus_pair(L @ U.T).new(mask=L.S).reduce_scalar().get(0)


def transitivity(G):
    numerator = total_triangles(G)
    if numerator == 0:
        return 0
    degrees = G.get_property("degrees-")
    denom = (degrees * (degrees - 1)).reduce().value
    return 6 * numerator / denom


def transitivity_directed(G):
    # XXX" is transitivity supposed to work on directed graphs like this?
    A, AT = G.get_properties("offdiag AT")
    numerator = plus_pair(A @ A.T).new(mask=A.S).reduce_scalar().get(0)
    if numerator == 0:
        return 0
    degrees = G.get_property("row_degrees-")
    denom = (degrees * (degrees - 1)).reduce().value
    return numerator / denom


def clustering(G, *, weighted=False, mask=None):
    tri = triangles(G, weighted=weighted, mask=mask)
    degrees = G.get_property("degrees-")
    denom = degrees * (degrees - 1)
    return (2 * tri / denom).new(name="clustering")


def clustering_directed(G, *, weighted=False, mask=None):
    # Can we apply the mask earlier in the computation?
    A, AT = G.get_properties("offdiag AT")
    if weighted:
        maxval = G.get_property("max_element-")
        A = unary.cbrt(A / maxval)
        AT = unary.cbrt(AT / maxval)
        semiring = plus_times
    else:
        semiring = plus_pair
    C = semiring(A @ A.T).new(mask=A.S)
    D = semiring(AT @ A.T).new(mask=A.S)
    E = semiring(AT @ AT.T).new(mask=A.S)
    if weighted:
        C *= A
        D *= A
        E *= A
    tri = (
        C.reduce_rowwise().new(mask=mask)
        + C.reduce_columnwise().new(mask=mask)
        + D.reduce_rowwise().new(mask=mask)
        + E.reduce_columnwise().new(mask=mask)
    )
    recip_degrees, total_degrees = G.get_properties("recip_degrees- total_degrees-", mask=mask)
    denom = total_degrees * (total_degrees - 1) - 2 * recip_degrees
    return (tri / denom).new(name="clustering")


def single_clustering(G, node, *, weighted=False):
    tri = single_triangle(G, node, weighted=weighted)
    if tri == 0:
        return 0
    index = G._key_to_id[node]
    # TODO: it would be nice if we could clean this up, but still be fast
    if "degrees-" in G._cache:
        degrees = G.get_property("degrees-").get(index)
    elif "degrees+" in G._cache:
        degrees = G.get_property("degrees+").get(index)
        if G.get_property("has_self_edges") and G._A.get(index, index) is not None:
            degrees -= 1
    else:
        row = G._A[index, :]
        degrees = row.nvals
        if G.get_property("has_self_edges") and row.get(index) is not None:
            degrees -= 1
    denom = degrees * (degrees - 1)
    return 2 * tri / denom


def single_clustering_directed(G, node, *, weighted=False):
    A = G.get_property("offdiag")
    index = G._key_to_id[node]
    if weighted:
        maxval = G.get_property("max_element-")
        A = unary.cbrt(A / maxval)
        semiring = plus_times
    else:
        semiring = plus_pair
    r = A[index, :]
    c = A[:, index]
    tris = []
    for x, y in [(c, c), (c, r), (r, c), (r, r)]:
        v = semiring(A @ x).new(mask=y.S)
        if weighted:
            v *= y
        tris.append(v.reduce().new())
    # Getting Python scalars are blocking operations, so we do them last
    tri = sum(t.get(0) for t in tris)
    if tri == 0:
        return 0
    total_degrees = c.nvals + r.nvals
    recip_degrees = binary.pair(c & r).nvals
    return tri / (total_degrees * (total_degrees - 1) - 2 * recip_degrees)


def average_clustering(G, *, count_zeros=True, weighted=False, mask=None):
    c = clustering(G, weighted=weighted, mask=mask)
    val = c.reduce().get(0)
    if not count_zeros:
        return val / c.nvals
    elif mask is not None:
        return val / mask.parent.nvals
    else:
        return val / c.size


def average_clustering_directed(G, *, count_zeros=True, weighted=False, mask=None):
    c = clustering_directed(G, weighted=weighted, mask=mask)
    val = c.reduce().get(0)
    if not count_zeros:
        return val / c.nvals
    elif mask is not None:
        return val / mask.parent.nvals
    else:
        return val / c.size


def single_square_clustering(G, idx):
    A, degrees = G.get_properties("A degrees+")  # TODO" how to handle self-edges?
    deg = degrees.get(idx, 0)
    if deg <= 1:
        return 0
    # P2 from https://arxiv.org/pdf/2007.11111.pdf; we'll also use it as scratch
    v = A[idx, :].new(name="v")
    p2 = plus_pair(v @ A).new(name="p2")
    del p2[idx]
    # Denominator is thought of as the total number of squares that could exist.
    # We use the definition from https://arxiv.org/pdf/0710.0117v1.pdf (equation 2).
    #
    # (1) Subtract 1 for each edge where u-w or w-u are connected (which would make triangles)
    denom = -plus_first(p2 @ v).get(0)
    # Numerator: number of squares
    # Based on https://arxiv.org/pdf/2007.11111.pdf (sigma_12, c_4)
    p2(binary.times) << p2 - 1
    squares = p2.reduce().get(0) // 2
    if squares == 0:
        return 0
    # (2) Subtract 1 for each u and 1 for each w for all combos: degrees * (degrees - 1)
    denom -= deg * (deg - 1)
    # (3) The main contribution to the denominator: degrees[u] + degrees[w] for each u-w combo.
    # This is the only positive term.
    denom += plus_times(v @ degrees).value * (deg - 1)
    # (4) Subtract the number of squares
    denom -= squares
    # And we're done!
    return squares / denom


def square_clustering(G, node_ids=None):
    # Warning: only tested on undirected graphs.
    # Also, it may use a lot of memory, because we compute `P2 = A @ A.T`
    #
    # Pseudocode:
    #   P2(~degrees.diag().S) = plus_pair(A @ A.T)
    #   tri = first(P2 & A).reduce_rowwise()
    #   squares = (P2 * (P2 - 1)).reduce_rowwise() / 2
    #   uw_count = degrees * (degrees - 1)
    #   uw_degrees = plus_times(A @ degrees) * (degrees - 1)
    #   square_clustering = squares / (uw_degrees - uw_count - tri - squares)
    #
    A, degrees = G.get_properties("A degrees+")  # TODO" how to handle self-edges?
    # P2 from https://arxiv.org/pdf/2007.11111.pdf; we'll also use it as scratch
    if node_ids is not None:
        v = Vector.from_coo(node_ids, True, size=degrees.size)
        Asubset = binary.second(v & A).new(name="A_subset")
    else:
        Asubset = A
        A = A.T
    D = degrees.diag(name="D")
    P2 = plus_pair(Asubset @ A).new(mask=~D.S, name="P2")

    # Denominator is thought of as the total number of squares that could exist.
    # We use the definition from https://arxiv.org/pdf/0710.0117v1.pdf (equation 2).
    #   denom = uw_degrees - uw_count - tri - squares
    #
    # (1) Subtract 1 for each edge where u-w or w-u are connected (i.e., triangles)
    #   tri = first(P2 & A).reduce_rowwise()
    D << binary.first(P2 & Asubset)
    neg_denom = D.reduce_rowwise().new(name="neg_denom")
    del D

    # Numerator: number of squares
    # Based on https://arxiv.org/pdf/2007.11111.pdf (sigma_12, c_4)
    #   squares = (P2 * (P2 - 1)).reduce_rowwise() / 2
    P2(binary.times) << P2 - 1
    squares = P2.reduce_rowwise().new(name="squares")
    del P2
    squares(squares.V, replace) << binary.cdiv(squares, 2)  # Drop zeros

    # (2) Subtract 1 for each u and 1 for each w for all combos: degrees * (degrees - 1)
    #   uw_count = degrees * (degrees - 1)
    denom = (degrees - 1).new(mask=squares.S, name="denom")
    neg_denom(binary.plus) << degrees * denom

    # (3) The main contribution to the denominator: degrees[u] + degrees[w] for each u-w combo.
    #   uw_degrees = plus_times(A @ degrees) * (degrees - 1)
    # denom(binary.times) << plus_times(A @ degrees)
    denom(binary.times, denom.S) << plus_times(Asubset @ degrees)

    # (4) Subtract the number of squares
    denom(binary.minus) << binary.plus(neg_denom & squares)

    # And we're done!  This result does not include 0s
    return (squares / denom).new(name="square_clustering")


def generalized_degree(G, *, mask=None):
    # Not benchmarked or optimized
    A = G.get_property("offdiag")
    Tri = Matrix(int, A.nrows, A.ncols, name="Tri")
    if mask is not None:
        if mask.structure and not mask.value:
            v_mask = mask.parent
        else:
            v_mask = mask.new()  # Not covered
        Tri << binary.pair(v_mask & A)  # Mask out rows
        Tri(Tri.S) << 0
    else:
        Tri(A.S) << 0
    Tri(Tri.S, binary.second) << plus_pair(Tri @ A.T)
    rows, cols, vals = Tri.to_coo()
    # The column index indicates the number of triangles an edge participates in.
    # The largest this can be is `A.ncols - 1`.  Values is count of edges.
    return Matrix.from_coo(
        rows,
        vals,
        np.ones(vals.size, dtype=int),
        dup_op=binary.plus,
        nrows=A.nrows,
        ncols=A.ncols - 1,
        name="generalized_degree",
    )


def single_generalized_degree(G, node):
    # Not benchmarked or optimized
    index = G._key_to_id[node]
    v = Vector(bool, len(G))
    v[index] = True
    return generalized_degree(G, mask=v.S)[index, :].new(name=f"generalized_degree_{index}")



from graphblas import Vector

from graphblas_algorithms.algorithms._helpers import is_converged, normalize
from graphblas_algorithms.algorithms.exceptions import (
    ConvergenceFailure,
    GraphBlasAlgorithmException,
    PointlessConcept,
)

__all__ = ["eigenvector_centrality"]


def eigenvector_centrality(G, max_iter=100, tol=1.0e-6, nstart=None, name="eigenvector_centrality"):
    N = len(G)
    if N == 0:
        raise PointlessConcept("cannot compute centrality for the null graph")
    x = Vector(float, N, name="x")
    if nstart is None:
        x << 1.0 / N
    else:
        x << nstart
        denom = x.reduce().get(0)  # why not use L2 norm?
        if denom == 0:
            raise GraphBlasAlgorithmException("initial vector cannot have all zero values")
        x *= 1.0 / denom

    # Power iteration: make up to max_iter iterations
    A = G._A
    xprev = Vector(float, N, name="x_prev")
    for _ in range(max_iter):
        xprev << x
        x += x @ A
        normalize(x, "L2")
        if is_converged(xprev, x, tol):  # sum(abs(xprev - x)) < N * tol
            x.name = name
            return x
    raise ConvergenceFailure(max_iter)



from . import exceptions
from .boundary import *
from .centrality import *
from .cluster import *
from .community import *
from .core import *
from .cuts import *
from .dag import *
from .dominating import *
from .isolate import *
from .link_analysis import *
from .reciprocity import *
from .regular import *
from .shortest_paths import *
from .simple_paths import *
from .smetric import *
from .structuralholes import *
from .tournament import *
from .triads import *



__all__ = ["is_isolate", "isolates", "number_of_isolates"]


def is_isolate(G, n):
    index = G._key_to_id[n]
    if G.is_directed():
        degrees = G.get_property("total_degrees+")
    else:
        degrees = G.get_property("degrees+")
    return index not in degrees


def isolates(G):
    if G.is_directed():
        degrees = G.get_property("total_degrees+")
    else:
        degrees = G.get_property("degrees+")
    return (~degrees.S).new(name="isolates")


def number_of_isolates(G):
    if G.is_directed():
        degrees = G.get_property("total_degrees+")
    else:
        degrees = G.get_property("degrees+")
    return degrees.size - degrees.nvals



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph

__all__ = []


def intra_community_edges(G, partition):
    G = to_graph(G)
    partition = [G.set_to_vector(block, ignore_extra=True) for block in partition]
    return algorithms.intra_community_edges(G, partition)


def inter_community_edges(G, partition):
    G = to_graph(G)
    partition = [G.set_to_vector(block, ignore_extra=True) for block in partition]
    return algorithms.inter_community_edges(G, partition)



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph
from graphblas_algorithms.classes.graph import to_undirected_graph
from graphblas_algorithms.utils import not_implemented_for

__all__ = ["is_regular", "is_k_regular"]


def is_regular(G):
    G = to_graph(G)
    return algorithms.is_regular(G)


@not_implemented_for("directed")
def is_k_regular(G, k):
    G = to_undirected_graph(G)
    return algorithms.is_k_regular(G, k)



from collections import defaultdict

import graphblas as gb
from graphblas import Matrix, Vector, select

import graphblas_algorithms as ga

from . import _utils
from ._caching import get_reduce_to_scalar, get_reduce_to_vector


def get_A(G, mask=None):
    """A"""
    return G._A


def get_AT(G, mask=None):
    """A.T"""
    A = G._A
    G._cache["AT"] = A
    return A


def get_offdiag(G, mask=None):
    """select.offdiag(A)"""
    A = G._A
    cache = G._cache
    if "offdiag" not in cache:
        if cache.get("has_self_edges") is False:
            cache["offdiag"] = A
        else:
            cache["offdiag"] = select.offdiag(A).new(name="offdiag")
    if "has_self_edges" not in cache:
        cache["has_self_edges"] = A.nvals > cache["offdiag"].nvals
    if not cache["has_self_edges"]:
        cache["offdiag"] = A
    return cache["offdiag"]


def get_Up(G, mask=None):
    """select.triu(A)"""
    A = G._A
    cache = G._cache
    if "U+" not in cache:
        if "U-" in cache and not G.get_property("has_self_edges"):
            cache["U+"] = cache["U-"]
        else:
            cache["U+"] = select.triu(A).new(name="U+")
    if "has_self_edges" not in cache:
        cache["has_self_edges"] = 2 * cache["U+"].nvals > A.nvals
    if not cache["has_self_edges"]:
        cache["U-"] = cache["U+"]
    return cache["U+"]


def get_Lp(G, mask=None):
    """select.tril(A)"""
    A = G._A
    cache = G._cache
    if "L+" not in cache:
        if "L-" in cache and not G.get_property("has_self_edges"):
            cache["L+"] = cache["L-"]
        else:
            cache["L+"] = select.tril(A).new(name="L+")
    if "has_self_edges" not in cache:
        cache["has_self_edges"] = 2 * cache["L+"].nvals > A.nvals
    if not cache["has_self_edges"]:
        cache["L-"] = cache["L+"]
    return cache["L+"]


def get_Um(G, mask=None):
    """select.triu(A, 1)"""
    A = G._A
    cache = G._cache
    if "U-" not in cache:
        if "U+" in cache:
            if G.get_property("has_self_edges"):
                cache["U-"] = select.triu(cache["U+"], 1).new(name="U-")
            else:
                cache["U-"] = cache["U+"]
        elif "offdiag" in cache:
            cache["U-"] = select.triu(cache["offdiag"], 1).new(name="U-")
        else:
            cache["U-"] = select.triu(A, 1).new(name="U-")
    if "has_self_edges" not in cache:
        cache["has_self_edges"] = 2 * cache["U-"].nvals < A.nvals
    if not cache["has_self_edges"]:
        cache["U+"] = cache["U-"]
    return cache["U-"]


def get_Lm(G, mask=None):
    """select.tril(A, -1)"""
    A = G._A
    cache = G._cache
    if "L-" not in cache:
        if "L+" in cache:
            if G.get_property("has_self_edges"):
                cache["L-"] = select.tril(cache["L+"], -1).new(name="L-")
            else:
                cache["L-"] = cache["L+"]
        elif "offdiag" in cache:
            cache["L-"] = select.tril(cache["offdiag"], -1).new(name="L-")
        else:
            cache["L-"] = select.tril(A, -1).new(name="L-")
    if "has_self_edges" not in cache:
        cache["has_self_edges"] = 2 * cache["L-"].nvals < A.nvals
    if not cache["has_self_edges"]:
        cache["L+"] = cache["L-"]
    return cache["L-"]


def get_diag(G, mask=None):
    """A.diag()"""
    A = G._A
    cache = G._cache
    if "diag" not in cache:
        if cache.get("has_self_edges") is False:
            cache["diag"] = Vector(A.dtype, size=A.nrows, name="diag")
        elif "U+" in cache:
            cache["diag"] = cache["U+"].diag(name="diag")
        elif "L+" in cache:
            cache["diag"] = cache["L+"].diag(name="diag")
        else:
            cache["diag"] = A.diag(name="diag")
    if "has_self_edges" not in cache:
        cache["has_self_edges"] = cache["diag"].nvals > 0
    return cache["diag"]


def has_self_edges(G, mask=None):
    """A.diag().nvals > 0"""
    A = G._A
    cache = G._cache
    if "has_self_edges" not in cache:
        if "L+" in cache:
            cache["has_self_edges"] = 2 * cache["L+"].nvals > A.nvals
        elif "L-" in cache:
            cache["has_self_edges"] = 2 * cache["L-"].nvals < A.nvals
        elif "U+" in cache:
            cache["has_self_edges"] = 2 * cache["U+"].nvals > A.nvals
        elif "U-" in cache:
            cache["has_self_edges"] = 2 * cache["U-"].nvals < A.nvals
        elif "offdiag" in cache:
            cache["has_self_edges"] = A.nvals > cache["offdiag"].nvals
        else:
            G.get_property("diag")
    return cache["has_self_edges"]


def to_undirected_graph(G, weight=None, dtype=None):
    # We should do some sanity checks here to ensure we're returning a valid undirected graph
    if isinstance(G, Graph):
        return G
    try:
        return Graph(G)
    except TypeError:
        pass

    try:
        import networkx as nx

        if isinstance(G, nx.Graph):
            return Graph.from_networkx(G, weight=weight, dtype=dtype)
    except ImportError:
        pass

    raise TypeError()


class AutoDict(dict):
    def __missing__(self, key):
        # Automatically compute keys such as "plus_rowwise-" and "max_element+"
        if key[-1] in {"-", "+"}:
            keybase = key[:-1]
            if keybase.endswith("_rowwise"):
                opname = keybase[: -len("_rowwise")]
                methodname = "reduce_rowwise"
            elif keybase.endswith("_columnwise"):
                opname = keybase[: -len("_columnwise")]
                methodname = "reduce_rowwise"
            elif keybase.endswith("_element"):
                opname = keybase[: -len("_element")]
                methodname = "reduce_scalar"
            else:
                raise KeyError(key)
            if methodname == "reduce_scalar":
                get_reduction = get_reduce_to_scalar(key, opname)
            else:
                get_reduction = get_reduce_to_vector(key, opname, methodname)
                self[f"{opname}_columnwise{key[-1]}"] = get_reduction
            self[key] = get_reduction
            return get_reduction
        raise KeyError(key)


class Graph:
    __networkx_plugin__ = "graphblas"

    # "-" properties ignore self-edges, "+" properties include self-edges
    # Ideally, we would have "max_rowwise+" come before "max_element+".
    _property_priority = defaultdict(
        lambda: Graph._property_priority["has_self_edges"] - 0.5,
        {
            key: i
            for i, key in enumerate(
                [
                    "A",
                    "AT",
                    "offdiag",
                    "U+",
                    "L+",
                    "U-",
                    "L-",
                    "diag",
                    "count_rowwise+",
                    "count_rowwise-",
                    "has_self_edges",
                ]
            )
        },
    )
    _get_property = AutoDict(
        {
            "A": get_A,
            "AT": get_AT,
            "offdiag": get_offdiag,
            "U+": get_Up,
            "L+": get_Lp,
            "U-": get_Um,
            "L-": get_Lm,
            "diag": get_diag,
            "has_self_edges": has_self_edges,
        }
    )
    _cache_aliases = {
        "degrees+": "count_rowwise+",
        "degrees-": "count_rowwise-",
        "row_degrees+": "count_rowwise+",
        "row_degrees-": "count_rowwise-",
        "column_degrees+": "count_rowwise+",
        "column_degrees-": "count_rowwise-",
    }
    graph_attr_dict_factory = dict

    def __init__(self, incoming_graph_data=None, *, key_to_id=None, **attr):
        if incoming_graph_data is not None:
            # Does not copy if A is a Matrix!
            A = gb.core.utils.ensure_type(incoming_graph_data, Matrix)
            if A.nrows != A.ncols:
                raise ValueError(f"Adjacency matrix must be square; got {A.nrows} x {A.ncols}")
        else:
            A = Matrix()
        self.graph_attr_dict_factory = self.graph_attr_dict_factory
        self.graph = self.graph_attr_dict_factory()  # dictionary for graph attributes
        self.graph.update(attr)

        # Graphblas-specific properties
        self._A = A
        if key_to_id is None:
            key_to_id = {i: i for i in range(A.nrows)}
        self._key_to_id = key_to_id
        self._id_to_key = None
        self._cache = {}

    # Graphblas-specific methods
    from_networkx = classmethod(_utils.from_networkx)
    id_to_key = property(_utils.id_to_key)
    get_property = _utils.get_property
    get_properties = _utils.get_properties
    dict_to_vector = _utils.dict_to_vector
    list_to_vector = _utils.list_to_vector
    list_to_mask = _utils.list_to_mask
    list_to_ids = _utils.list_to_ids
    list_to_keys = _utils.list_to_keys
    matrix_to_dicts = _utils.matrix_to_dicts
    set_to_vector = _utils.set_to_vector
    to_networkx = _utils.to_networkx
    vector_to_dict = _utils.vector_to_dict
    vector_to_nodemap = _utils.vector_to_nodemap
    vector_to_nodeset = _utils.vector_to_nodeset
    vector_to_set = _utils.vector_to_set
    _cacheit = _utils._cacheit

    # NetworkX methods
    def to_directed_class(self):
        return ga.DiGraph

    def to_undirected_class(self):
        return Graph

    @property
    def name(self):
        return self.graph.get("name", "")

    @name.setter
    def name(self, s):
        self._A.name = s
        self.graph["name"] = s

    @property
    def matrix(self):
        return self._A

    def __iter__(self):
        return iter(self._key_to_id)

    def __contains__(self, n):
        try:
            return n in self._key_to_id
        except TypeError:
            return False

    def __len__(self):
        return self._A.nrows

    def number_of_nodes(self):
        return self._A.nrows

    def order(self):
        return self._A.nrows

    def is_multigraph(self):
        return False

    def is_directed(self):
        return False


class MultiGraph(Graph):
    def is_multigraph(self):
        return True


__all__ = ["Graph", "MultiGraph"]



from collections import defaultdict

import graphblas as gb
from graphblas import Matrix, Vector, binary, replace, select, unary

import graphblas_algorithms as ga

from . import _utils
from ._caching import get_reduce_to_scalar, get_reduce_to_vector
from .graph import Graph


def get_A(G, mask=None):
    """A"""
    return G._A


def get_AT(G, mask=None):
    """A.T"""
    A = G._A
    cache = G._cache
    if "AT" not in cache:
        cache["AT"] = A.T.new()
    return cache["AT"]


def get_offdiag(G, mask=None):
    """select.offdiag(A)"""
    A = G._A
    cache = G._cache
    if "offdiag" not in cache:
        if cache.get("has_self_edges") is False:
            cache["offdiag"] = A
        else:
            cache["offdiag"] = select.offdiag(A).new(name="offdiag")
    if "has_self_edges" not in cache:
        cache["has_self_edges"] = A.nvals > cache["offdiag"].nvals
    if not cache["has_self_edges"]:
        cache["offdiag"] = A
    return cache["offdiag"]


def get_Up(G, mask=None):
    """select.triu(A)"""
    A = G._A
    cache = G._cache
    if "U+" not in cache:
        if "U-" in cache and not G.get_property("has_self_edges"):
            cache["U+"] = cache["U-"]
        else:
            cache["U+"] = select.triu(A).new(name="U+")
    if "has_self_edges" not in cache:
        if "U-" in cache:
            cache["has_self_edges"] = cache["U+"].nvals > cache["U-"].nvals
        elif "L+" in cache:
            cache["has_self_edges"] = cache["U+"].nvals + cache["L+"].nvals > A.nvals
    if cache.get("has_self_edges") is False:
        cache["U-"] = cache["U+"]
    return cache["U+"]


def get_Lp(G, mask=None):
    """select.tril(A)"""
    A = G._A
    cache = G._cache
    if "L+" not in cache:
        if "L-" in cache and not G.get_property("has_self_edges"):
            cache["L+"] = cache["L-"]
        else:
            cache["L+"] = select.tril(A).new(name="L+")
    if "has_self_edges" not in cache:
        if "L-" in cache:
            cache["has_self_edges"] = cache["L+"].nvals > cache["L-"].nvals
        elif "U+" in cache:
            cache["has_self_edges"] = cache["L+"].nvals + cache["U+"].nvals > A.nvals
    if cache.get("has_self_edges") is False:
        cache["L-"] = cache["L+"]
    return cache["L+"]


def get_Um(G, mask=None):
    """select.triu(A, 1)"""
    A = G._A
    cache = G._cache
    if "U-" not in cache:
        if "U+" in cache:
            if cache.get("has_self_edges") is False:
                cache["U-"] = cache["U+"]
            else:
                cache["U-"] = select.triu(cache["U+"], 1).new(name="U-")
        elif "offdiag" in cache:
            cache["U-"] = select.triu(cache["offdiag"], 1).new(name="U-")
        else:
            cache["U-"] = select.triu(A, 1).new(name="U-")
    if "has_self_edges" not in cache:
        if "U+" in cache:
            cache["has_self_edges"] = cache["U-"].nvals < cache["U+"].nvals
        elif "L-" in cache:
            cache["has_self_edges"] = cache["U-"].nvals + cache["L-"].nvals < A.nvals
    if cache.get("has_self_edges") is False:
        cache["U+"] = cache["U-"]
    return cache["U-"]


def get_Lm(G, mask=None):
    """select.tril(A, -1)"""
    A = G._A
    cache = G._cache
    if "L-" not in cache:
        if "L+" in cache:
            if cache.get("has_self_edges") is False:
                cache["L-"] = cache["L+"]
            else:
                cache["L-"] = select.tril(cache["L+"], -1).new(name="L-")
        elif "offdiag" in cache:
            cache["L-"] = select.tril(cache["offdiag"], -1).new(name="L-")
        else:
            cache["L-"] = select.tril(A, -1).new(name="L-")
    if "has_self_edges" not in cache:
        if "L+" in cache:
            cache["has_self_edges"] = cache["L-"].nvals < cache["L+"].nvals
        elif "U-" in cache:
            cache["has_self_edges"] = cache["L-"].nvals + cache["U-"].nvals < A.nvals
    if cache.get("has_self_edges") is False:
        cache["L+"] = cache["L-"]
    return cache["L-"]


def get_diag(G, mask=None):
    """select.diag(A)"""
    A = G._A
    cache = G._cache
    if "diag" not in cache:
        if cache.get("has_self_edges") is False:
            cache["diag"] = Vector(A.dtype, size=A.nrows, name="diag")
        elif "U+" in cache:
            cache["diag"] = cache["U+"].diag(name="diag")
        elif "L+" in cache:
            cache["diag"] = cache["L+"].diag(name="diag")
        else:
            cache["diag"] = A.diag(name="diag")
    if "has_self_edges" not in cache:
        cache["has_self_edges"] = cache["diag"].nvals > 0
    if mask is not None:
        return cache["diag"].dup(mask=mask)
    return cache["diag"]


def get_recip_degreesp(G, mask=None):
    """pair(A & A.T).reduce_rowwise()"""
    A = G._A
    cache = G._cache
    if "AT" in cache:
        AT = cache["AT"]
    else:
        AT = A.T
    if mask is not None:
        if "recip_degrees+" in cache:
            return cache["recip_degrees+"].dup(mask=mask)
        elif cache.get("has_self_edges") is False and "recip_degrees-" in cache:
            cache["recip_degrees+"] = cache["recip_degrees-"]
            return cache["recip_degrees-"].dup(mask=mask)
        elif "recip_degrees-" in cache and "diag" in cache:
            return (unary.one(cache["diag"]) + cache["recip_degrees-"]).new(
                mask=mask, name="recip_degrees+"
            )
        elif "recip_degrees-" in cache and not G.get_property("has_self_edges"):
            return cache["recip_degrees-"].dup(mask=mask)
        else:
            return binary.pair(A & AT).reduce_rowwise().new(mask=mask, name="recip_degrees+")
    if "recip_degrees+" not in cache:
        if cache.get("has_self_edges") is False and "recip_degrees-" in cache:
            cache["recip_degrees+"] = cache["recip_degrees-"]
        elif "recip_degrees-" in cache and "diag" in cache:
            cache["recip_degrees+"] = (unary.one(cache["diag"]) + cache["recip_degrees-"]).new(
                name="recip_degrees+"
            )
        elif "recip_degrees-" in cache and not G.get_property("has_self_edges"):
            cache["recip_degrees+"] = cache["recip_degrees-"]
        else:
            cache["recip_degrees+"] = (
                binary.pair(A & AT).reduce_rowwise().new(name="recip_degrees+")
            )
    if (
        "has_self_edges" not in cache
        and "recip_degrees-" in cache
        and cache["recip_degrees-"].nvals != cache["recip_degrees+"].nvals
    ):
        cache["has_self_edges"] = True
    elif cache.get("has_self_edges") is False:
        cache["recip_degrees-"] = cache["recip_degrees+"]
    return cache["recip_degrees+"]


def get_recip_degreesm(G, mask=None):
    """C = select.offdiag(A) ; pair(C & C.T).reduce_rowwise()"""
    A = G._A
    cache = G._cache
    if "AT" in cache:
        AT = cache["AT"]
    elif "offdiag" in cache:
        AT = cache["offdiag"].T
    else:
        AT = A.T
    if mask is not None:
        if "recip_degrees-" in cache:
            return cache["recip_degrees-"].dup(mask=mask)
        elif cache.get("has_self_edges") is False and "recip_degrees+" in cache:
            cache["recip_degrees-"] = cache["recip_degrees+"]
            return cache["recip_degrees-"].dup(mask=mask)
        elif "recip_degrees+" in cache and "diag" in cache:
            rv = binary.minus(cache["recip_degrees+"] | unary.one(cache["diag"])).new(
                mask=mask, name="recip_degrees-"
            )
            rv(rv.V, replace) << rv  # drop 0s
            return rv
        elif not G.get_property("has_self_edges"):
            return G.get_property("recip_degrees+", mask=mask)
        elif "offdiag" in cache:
            return (
                binary.pair(cache["offdiag"] & AT)
                .reduce_rowwise()
                .new(mask=mask, name="recip_degrees-")
            )
        elif "L-" in cache and "U-" in cache:
            return (
                binary.pair(cache["L-"] & AT).reduce_rowwise().new(mask=mask)
                + binary.pair(cache["U-"] & AT).reduce_rowwise().new(mask=mask)
            ).new(name="recip_degrees-")
        else:
            diag = G.get_property("diag", mask=mask)
            overlap = binary.pair(A & AT).reduce_rowwise().new(mask=mask)
            rv = binary.minus(overlap | unary.one(diag)).new(name="recip_degrees-")
            rv(rv.V, replace) << rv  # drop 0s
            return rv
    if "recip_degrees-" not in cache:
        if cache.get("has_self_edges") is False and "recip_degrees+" in cache:
            cache["recip_degrees-"] = cache["recip_degrees+"]
        elif "recip_degrees+" in cache and "diag" in cache:
            rv = binary.minus(cache["recip_degrees+"] | unary.one(cache["diag"])).new(
                name="recip_degrees-"
            )
            rv(rv.V, replace) << rv  # drop 0s
            cache["recip_degrees-"] = rv
        elif not G.get_property("has_self_edges"):
            cache["recip_degrees-"] = G.get_property("recip_degrees+")
        elif "offdiag" in cache:
            cache["recip_degrees-"] = (
                binary.pair(cache["offdiag"] & AT).reduce_rowwise().new(name="recip_degrees-")
            )
        elif "L-" in cache and "U-" in cache:
            cache["recip_degrees-"] = (
                binary.pair(cache["L-"] & AT).reduce_rowwise().new()
                + binary.pair(cache["U-"] & AT).reduce_rowwise().new()
            ).new(name="recip_degrees-")
        else:
            diag = G.get_property("diag")
            overlap = binary.pair(A & AT).reduce_rowwise().new()
            rv = binary.minus(overlap | unary.one(diag)).new(name="recip_degrees-")
            rv(rv.V, replace) << rv  # drop 0s
            cache["recip_degrees-"] = rv
    if (
        "has_self_edges" not in cache
        and "recip_degrees+" in cache
        and cache["recip_degrees-"].nvals != cache["recip_degrees+"].nvals
    ):
        cache["has_self_edges"] = True
    elif cache.get("has_self_edges") is False:
        cache["recip_degrees+"] = cache["recip_degrees-"]
    return cache["recip_degrees-"]


def get_total_degreesp(G, mask=None):
    """A.reduce_rowwise(agg.count) + A.reduce_columnwise(agg.count)"""
    cache = G._cache
    if mask is not None:
        if "total_degrees+" in cache:
            return cache["total_degrees+"].dup(mask=mask)
        elif cache.get("has_self_edges") is False and "total_degrees-" in cache:
            cache["total_degrees+"] = cache["total_degrees-"]
            return cache["total_degrees+"].dup(mask=mask)
        else:
            return (
                G.get_property("row_degrees+", mask=mask)
                + G.get_property("column_degrees+", mask=mask)
            ).new(name="total_degrees+")
    if "total_degrees+" not in cache:
        if cache.get("has_self_edges") is False and "total_degrees-" in cache:
            cache["total_degrees+"] = cache["total_degrees-"]
        else:
            cache["total_degrees+"] = (
                G.get_property("row_degrees+") + G.get_property("column_degrees+")
            ).new(name="total_degrees+")
    if (
        "has_self_edges" not in cache
        and "total_degrees-" in cache
        and cache["total_degrees-"].nvals != cache["total_degrees+"].nvals
    ):
        cache["has_self_edges"] = True
    elif cache.get("has_self_edges") is False:
        cache["total_degrees-"] = cache["total_degrees+"]
    return cache["total_degrees+"]


def get_total_degreesm(G, mask=None):
    """C = select.offdiag(A) ; C.reduce_rowwise(agg.count) + C.reduce_columnwise(agg.count)"""
    cache = G._cache
    if mask is not None:
        if "total_degrees-" in cache:
            return cache["total_degrees-"].dup(mask=mask)
        elif cache.get("has_self_edges") is False and "total_degrees+" in cache:
            cache["total_degrees-"] = cache["total_degrees+"]
            return cache["total_degrees-"].dup(mask=mask)
        else:
            return (
                G.get_property("row_degrees-", mask=mask)
                + G.get_property("column_degrees-", mask=mask)
            ).new(name="total_degrees-")
    if "total_degrees-" not in cache:
        if cache.get("has_self_edges") is False and "total_degrees+" in cache:
            cache["total_degrees-"] = cache["total_degrees+"]
        else:
            cache["total_degrees-"] = (
                G.get_property("row_degrees-") + G.get_property("column_degrees-")
            ).new(name="total_degrees-")
    if (
        "has_self_edges" not in cache
        and "total_degrees+" in cache
        and cache["total_degrees-"].nvals != cache["total_degrees+"].nvals
    ):
        cache["has_self_edges"] = True
    elif cache.get("has_self_edges") is False:
        cache["total_degrees+"] = cache["total_degrees-"]
    return cache["total_degrees-"]


def get_total_recipp(G, mask=None):
    """pair(A & A.T).reduce_scalar()"""
    A = G._A
    cache = G._cache
    if "total_recip+" not in cache:
        if "total_recip-" in cache and cache.get("has_self_edges") is False:
            cache["total_recip+"] = cache["total_recip-"]
        elif "recip_degrees+" in cache:
            cache["total_recip+"] = cache["recip_degrees+"].reduce().get(0)
        else:
            if "AT" in cache:
                AT = cache["AT"]
            else:
                AT = A.T
            cache["total_recip+"] = binary.pair(A & AT).reduce_scalar().get(0)
    if "has_self_edges" not in cache and "total_recip-" in cache:
        cache["has_self_edges"] = cache["total_recip+"] > cache["total_recip-"]
    if cache.get("has_self_edges") is False:
        cache["total_recip-"] = cache["total_recip+"]
    return cache["total_recip+"]


def get_total_recipm(G, mask=None):
    """C = select.offdiag(A) ; pair(C & C.T).reduce_scalar()"""
    cache = G._cache
    if "total_recip-" not in cache:
        if "total_recip+" in cache and cache.get("has_self_edges") is False:
            cache["total_recip-"] = cache["total_recip+"]
        else:
            cache["total_recip-"] = G.get_property("recip_degrees-").reduce().get(0)
    if "has_self_edges" not in cache and "total_recip+" in cache:
        cache["has_self_edges"] = cache["total_recip+"] > cache["total_recip-"]
    if cache.get("has_self_edges") is False:
        cache["total_recip+"] = cache["total_recip-"]
    return cache["total_recip-"]


def has_self_edges(G, mask=None):
    """A.diag().nvals > 0"""
    A = G._A
    cache = G._cache
    if "has_self_edges" not in cache:
        if "offdiag" in cache:
            cache["has_self_edges"] = A.nvals > cache["offdiag"].nvals
        elif "L+" in cache and ("L-" in cache or "U+" in cache):
            if "L-" in cache:
                cache["has_self_edges"] = cache["L-"].nvals < cache["L+"].nvals
            else:
                cache["has_self_edges"] = cache["L+"].nvals + cache["U+"].nvals > A.nvals
        elif "U-" in cache and ("U+" in cache or "L-" in cache):
            if "U+" in cache:
                cache["has_self_edges"] = cache["U-"].nvals < cache["U+"].nvals
            else:
                cache["has_self_edges"] = cache["U-"].nvals + cache["L-"].nvals < A.nvals
        elif "total_recip-" in cache and "total_recip+" in cache:
            cache["has_self_edges"] = cache["total_recip+"] > cache["total_recip-"]
        elif "row_degrees-" in cache and "row_degrees+" in cache:
            cache["has_self_edges"] = not cache["row_degrees-"].isequal(cache["row_degrees+"])
        elif "column_degrees-" in cache and "column_degrees+" in cache:
            cache["has_self_edges"] = not cache["column_degrees-"].isequal(cache["column_degrees+"])
        elif "total_degrees-" in cache and "total_degrees+" in cache:
            cache["has_self_edges"] = not cache["total_degrees-"].isequal(cache["total_degrees+"])
        elif "recip_degrees-" in cache and "recip_degrees+" in cache:
            cache["has_self_edges"] = not cache["recip_degrees-"].isequal(cache["recip_degrees+"])
        elif "row_degrees-" in cache:
            cache["has_self_edges"] = cache["row_degrees-"].reduce().get(0) < A.nvals
        elif "column_degrees-" in cache:
            cache["has_self_edges"] = cache["column_degrees-"].reduce().get(0) < A.nvals
        elif "total_degrees-" in cache:
            cache["has_self_edges"] = cache["total_degrees-"].reduce().get(0) < 2 * A.nvals
        elif "total_degrees+" in cache:
            cache["has_self_edges"] = cache["total_degrees+"].reduce().get(0) > 2 * A.nvals
        else:
            G.get_property("diag")
    return cache["has_self_edges"]


def to_directed_graph(G, weight=None, dtype=None):
    # We should do some sanity checks here to ensure we're returning a valid directed graph
    if isinstance(G, DiGraph):
        return G
    try:
        return DiGraph(G)
    except TypeError:
        pass

    try:
        import networkx as nx

        if isinstance(G, nx.DiGraph):
            return DiGraph.from_networkx(G, weight=weight, dtype=dtype)
    except ImportError:
        pass

    raise TypeError()


def to_graph(G, weight=None, dtype=None):
    if isinstance(G, (DiGraph, ga.Graph)):
        return G
    try:
        # Should we check if it can be undirected?
        return DiGraph(G)
    except TypeError:
        pass

    try:
        import networkx as nx

        if isinstance(G, nx.DiGraph):
            return DiGraph.from_networkx(G, weight=weight, dtype=dtype)
        if isinstance(G, nx.Graph):
            return ga.Graph.from_networkx(G, weight=weight, dtype=dtype)
    except ImportError:
        pass

    raise TypeError()


class AutoDict(dict):
    def __missing__(self, key):
        # Automatically compute keys such as "plus_rowwise-" and "max_element+"
        if key[-1] in {"-", "+"}:
            keybase = key[:-1]
            if keybase.endswith("_rowwise"):
                opname = keybase[: -len("_rowwise")]
                methodname = "reduce_rowwise"
            elif keybase.endswith("_columnwise"):
                opname = keybase[: -len("_columnwise")]
                methodname = "reduce_columnwise"
            elif keybase.endswith("_element"):
                opname = keybase[: -len("_element")]
                methodname = "reduce_scalar"
            else:
                raise KeyError(key)
            if methodname == "reduce_scalar":
                get_reduction = get_reduce_to_scalar(key, opname)
            else:
                get_reduction = get_reduce_to_vector(key, opname, methodname)
            self[key] = get_reduction
            return get_reduction
        raise KeyError(key)


class DiGraph(Graph):
    __networkx_plugin__ = "graphblas"

    # "-" properties ignore self-edges, "+" properties include self-edges
    # Ideally, we would have "max_rowwise+" come before "max_element+".
    _property_priority = defaultdict(
        lambda: DiGraph._property_priority["has_self_edges"] - 0.5,
        {
            key: i
            for i, key in enumerate(
                [
                    "A",
                    "AT",
                    "offdiag",
                    "U+",
                    "L+",
                    "U-",
                    "L-",
                    "diag",
                    "count_rowwise+",  # row_degrees
                    "count_columnwise+",  # column_degrees
                    "count_rowwise-",
                    "count_columnwise-",
                    "recip_degrees+",
                    "recip_degrees-",
                    "total_degrees+",
                    "total_degrees-",
                    "total_recip+",  # scalar; I don't like this name
                    "total_recip-",  # scalar; I don't like this name
                    "has_self_edges",
                ]
            )
        },
    )
    _get_property = AutoDict(
        {
            "A": get_A,
            "AT": get_AT,
            "offdiag": get_offdiag,
            "U+": get_Up,
            "L+": get_Lp,
            "U-": get_Um,
            "L-": get_Lm,
            "diag": get_diag,
            "recip_degrees+": get_recip_degreesp,
            "recip_degrees-": get_recip_degreesm,
            "total_degrees+": get_total_degreesp,
            "total_degrees-": get_total_degreesm,
            "total_recip+": get_total_recipp,
            "total_recip-": get_total_recipm,
            "has_self_edges": has_self_edges,
        }
    )
    _cache_aliases = {
        "row_degrees+": "count_rowwise+",
        "column_degrees+": "count_columnwise+",
        "row_degrees-": "count_rowwise-",
        "column_degrees-": "count_columnwise-",
    }
    graph_attr_dict_factory = dict

    def __init__(self, incoming_graph_data=None, *, key_to_id=None, **attr):
        if incoming_graph_data is not None:
            # Does not copy if A is a Matrix!
            A = gb.core.utils.ensure_type(incoming_graph_data, Matrix)
            if A.nrows != A.ncols:
                raise ValueError(f"Adjacency matrix must be square; got {A.nrows} x {A.ncols}")
        else:
            A = Matrix()
        self.graph_attr_dict_factory = self.graph_attr_dict_factory
        self.graph = self.graph_attr_dict_factory()  # dictionary for graph attributes
        self.graph.update(attr)

        # Graphblas-specific properties
        self._A = A
        if key_to_id is None:
            key_to_id = {i: i for i in range(A.nrows)}
        self._key_to_id = key_to_id
        self._id_to_key = None
        self._cache = {}

    # Graphblas-specific methods
    from_networkx = classmethod(_utils.from_networkx)
    id_to_key = property(_utils.id_to_key)
    get_property = _utils.get_property
    get_properties = _utils.get_properties
    dict_to_vector = _utils.dict_to_vector
    list_to_vector = _utils.list_to_vector
    list_to_mask = _utils.list_to_mask
    list_to_ids = _utils.list_to_ids
    matrix_to_dicts = _utils.matrix_to_dicts
    set_to_vector = _utils.set_to_vector
    to_networkx = _utils.to_networkx
    vector_to_dict = _utils.vector_to_dict
    vector_to_nodemap = _utils.vector_to_nodemap
    vector_to_nodeset = _utils.vector_to_nodeset
    vector_to_set = _utils.vector_to_set
    _cacheit = _utils._cacheit

    # NetworkX methods
    def to_directed_class(self):
        return DiGraph

    def to_undirected_class(self):
        return ga.Graph

    @property
    def name(self):
        return self.graph.get("name", "")

    @name.setter
    def name(self, s):
        self._A.name = s
        self.graph["name"] = s

    @property
    def matrix(self):
        return self._A

    def __iter__(self):
        return iter(self._key_to_id)

    def __contains__(self, n):
        try:
            return n in self._key_to_id
        except TypeError:
            return False

    def __len__(self):
        return self._A.nrows

    def number_of_nodes(self):
        return self._A.nrows

    def order(self):
        return self._A.nrows

    def is_multigraph(self):
        return False

    def is_directed(self):
        return True


class MultiDiGraph(DiGraph):
    def is_multigraph(self):
        return True


__all__ = ["DiGraph", "MultiDiGraph"]



from graphblas import Vector

__all__ = ["degree_centrality", "in_degree_centrality", "out_degree_centrality"]


def _degree_centrality(G, degrees, name):
    N = len(G)
    rv = Vector(float, size=N, name=name)
    if N <= 1:
        rv << 1
    else:
        s = 1 / (N - 1)
        rv << s * degrees
    return rv


def degree_centrality(G, *, name="degree_centrality"):
    if G.is_directed():
        degrees = G.get_property("total_degrees+")
    else:
        degrees = G.get_property("degrees+")
    return _degree_centrality(G, degrees, name)


def in_degree_centrality(G, *, name="in_degree_centrality"):
    degrees = G.get_property("column_degrees+")
    return _degree_centrality(G, degrees, name)


def out_degree_centrality(G, *, name="out_degree_centrality"):
    degrees = G.get_property("row_degrees+")
    return _degree_centrality(G, degrees, name)



from graphblas import io

from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_directed_graph
from graphblas_algorithms.utils import not_implemented_for

from .simple_paths import is_simple_path as is_path  # noqa

__all__ = ["is_tournament", "score_sequence", "tournament_matrix"]


@not_implemented_for("undirected")
@not_implemented_for("multigraph")
def is_tournament(G):
    G = to_directed_graph(G)
    return algorithms.is_tournament(G)


@not_implemented_for("undirected")
@not_implemented_for("multigraph")
def score_sequence(G):
    G = to_directed_graph(G)
    # TODO: can we return a different, more native object?
    return algorithms.score_sequence(G).tolist()


@not_implemented_for("undirected")
@not_implemented_for("multigraph")
def tournament_matrix(G):
    G = to_directed_graph(G)
    T = algorithms.tournament_matrix(G)
    # TODO: can we return a different, more native object?
    return io.to_scipy_sparse(T)



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph
from graphblas_algorithms.utils import not_implemented_for

from ..exception import NetworkXError, PowerIterationFailedConvergence

__all__ = ["katz_centrality"]


@not_implemented_for("multigraph")
def katz_centrality(
    G,
    alpha=0.1,
    beta=1.0,
    max_iter=1000,
    tol=1.0e-6,
    nstart=None,
    normalized=True,
    weight=None,
):
    G = to_graph(G, weight=weight, dtype=float)
    if len(G) == 0:
        return {}
    x = G.dict_to_vector(nstart, dtype=float, name="nstart")
    try:
        b = float(beta)
    except (TypeError, ValueError):
        try:
            b = G.dict_to_vector(beta, dtype=float, name="beta")
        except (TypeError, ValueError, AttributeError) as e:
            raise NetworkXError(*e.args) from e
    try:
        result = algorithms.katz_centrality(
            G, alpha=alpha, beta=b, max_iter=max_iter, tol=tol, nstart=x, normalized=normalized
        )
    except algorithms.exceptions.ConvergenceFailure as e:
        raise PowerIterationFailedConvergence(*e.args) from e
    except algorithms.exceptions.GraphBlasAlgorithmException as e:
        raise NetworkXError(*e.args) from e
    return G.vector_to_nodemap(result)



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph

from .exception import NetworkXError

__all__ = ["s_metric"]


def s_metric(G, normalized=True):
    if normalized:
        raise NetworkXError("Normalization not implemented")
    G = to_graph(G)
    return algorithms.s_metric(G)



from graphblas import monoid

from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph
from graphblas_algorithms.classes.graph import to_undirected_graph
from graphblas_algorithms.classes.nodemap import VectorNodeMap
from graphblas_algorithms.utils import not_implemented_for

__all__ = [
    "triangles",
    "transitivity",
    "clustering",
    "average_clustering",
    "square_clustering",
    "generalized_degree",
]


@not_implemented_for("directed")
def triangles(G, nodes=None):
    G = to_undirected_graph(G, dtype=bool)
    if len(G) == 0:
        return {}
    if nodes in G:
        return algorithms.single_triangle(G, nodes)
    mask = G.list_to_mask(nodes)
    result = algorithms.triangles(G, mask=mask)
    return G.vector_to_nodemap(result, mask=mask, fillvalue=0)


def transitivity(G):
    G = to_graph(G, dtype=bool)  # directed or undirected
    if len(G) == 0:
        return 0
    if G.is_directed():
        func = algorithms.transitivity_directed
    else:
        func = algorithms.transitivity
    return G._cacheit("transitivity", func, G)


def clustering(G, nodes=None, weight=None):
    G = to_graph(G, weight=weight)  # to directed or undirected
    if len(G) == 0:
        return {}
    weighted = weight is not None
    if nodes in G:
        if G.is_directed():
            return algorithms.single_clustering_directed(G, nodes, weighted=weighted)
        else:
            return algorithms.single_clustering(G, nodes, weighted=weighted)
    mask = G.list_to_mask(nodes)
    if G.is_directed():
        result = algorithms.clustering_directed(G, weighted=weighted, mask=mask)
    else:
        result = algorithms.clustering(G, weighted=weighted, mask=mask)
    return G.vector_to_nodemap(result, mask=mask, fillvalue=0.0)


def average_clustering(G, nodes=None, weight=None, count_zeros=True):
    G = to_graph(G, weight=weight)  # to directed or undirected
    if len(G) == 0:
        raise ZeroDivisionError()
    weighted = weight is not None
    mask = G.list_to_mask(nodes)
    if G.is_directed():
        func = algorithms.average_clustering_directed
    else:
        func = algorithms.average_clustering
    if mask is None:
        return G._cacheit(
            f"average_clustering(count_zeros={count_zeros})",
            func,
            G,
            weighted=weighted,
            count_zeros=count_zeros,
        )
    else:
        return func(G, weighted=weighted, count_zeros=count_zeros, mask=mask)


def _split(L, k):
    """Split a list into approximately-equal parts"""
    N = len(L)
    start = 0
    for i in range(1, k):
        stop = (N * i + k - 1) // k
        if stop != start:
            yield L[start:stop]
            start = stop
    if stop != N:
        yield L[stop:]


# TODO: should this move into algorithms?
def _square_clustering_split(G, node_ids=None, *, nsplits):
    if node_ids is None:
        node_ids = G._A.reduce_rowwise(monoid.any).to_coo()[0]
    result = None
    for chunk_ids in _split(node_ids, nsplits):
        res = algorithms.square_clustering(G, chunk_ids)
        if result is None:
            result = res
        else:
            result << monoid.any(result | res)
    return result


def square_clustering(G, nodes=None, *, nsplits=None):
    G = to_undirected_graph(G)
    if len(G) == 0:
        return {}
    elif nodes is None:
        # Should we use this one for subsets of nodes as well?
        if nsplits is None:
            result = algorithms.square_clustering(G)
        else:
            result = _square_clustering_split(G, nsplits=nsplits)
        return G.vector_to_nodemap(result, fillvalue=0)
    elif nodes in G:
        idx = G._key_to_id[nodes]
        return algorithms.single_square_clustering(G, idx)
    else:
        ids = G.list_to_ids(nodes)
        if nsplits is None:
            result = algorithms.square_clustering(G, ids)
        else:
            result = _square_clustering_split(G, ids, nsplits=nsplits)
        return G.vector_to_nodemap(result)


@not_implemented_for("directed")
def generalized_degree(G, nodes=None):
    G = to_undirected_graph(G)
    if len(G) == 0:
        return {}
    if nodes in G:
        result = algorithms.single_generalized_degree(G, nodes)
        return G.vector_to_nodemap(result)
    mask = G.list_to_mask(nodes)
    result = algorithms.generalized_degree(G, mask=mask)
    rv = VectorNodeMap(result, key_to_id=G._key_to_id)
    rv._id_to_key = G._id_to_key
    return rv



from graphblas import Vector, replace
from graphblas.semiring import lor_pair

__all__ = ["has_path"]


def has_path(G, source, target):
    # Perform bidirectional BFS from source to target and target to source
    src = G._key_to_id[source]
    dst = G._key_to_id[target]
    if src == dst:
        return True
    A = G._A
    q_src = Vector.from_coo(src, True, size=A.nrows, name="q_src")
    seen_src = q_src.dup(name="seen_src")
    q_dst = Vector.from_coo(dst, True, size=A.nrows, name="q_dst")
    seen_dst = q_dst.dup(name="seen_dst")
    for _ in range(A.nrows // 2):
        q_src(~seen_src.S, replace) << lor_pair(q_src @ A)
        if q_src.nvals == 0:
            return False
        if lor_pair(q_src @ q_dst):
            return True

        q_dst(~seen_dst.S, replace) << lor_pair(A @ q_dst)
        if q_dst.nvals == 0:
            return False
        if lor_pair(q_src @ q_dst):
            return True

        seen_src(q_src.S) << True
        seen_dst(q_dst.S) << True
    return False



from graphblas.semiring import lor_pair

__all__ = ["is_dominating_set"]


def is_dominating_set(G, nbunch):
    nbrs = lor_pair(nbunch @ G._A).new(mask=~nbunch.S)  # A or A.T?
    return nbrs.size - nbunch.nvals - nbrs.nvals == 0



class GraphBlasAlgorithmException(Exception):
    pass


class ConvergenceFailure(GraphBlasAlgorithmException):
    pass


class EmptyGraphError(GraphBlasAlgorithmException):
    pass


class PointlessConcept(GraphBlasAlgorithmException):
    pass



from graphblas import binary

__all__ = ["s_metric"]


def s_metric(G):
    if G.is_directed():
        degrees = G.get_property("total_degrees+")
    else:
        degrees = G.get_property("degrees+")
    return (binary.first(degrees & G._A) @ degrees).reduce().get(0) / 2
    # Alternatives
    # return (degrees @ binary.second(G._A & degrees)).reduce().get(0) / 2
    # return degrees.outer(degrees).new(mask=G._A.S).reduce_scalar().get(0) / 2



from graphblas import Vector, replace
from graphblas.semiring import lor_pair

__all__ = ["descendants", "ancestors"]


# Push-pull optimization is possible, but annoying to implement
def descendants(G, source):
    if source not in G._key_to_id:
        raise KeyError(f"The node {source} is not in the graph")
    index = G._key_to_id[source]
    A = G._A
    q = Vector.from_coo(index, True, size=A.nrows, name="q")
    rv = q.dup(name="descendants")
    for _ in range(A.nrows):
        q(~rv.S, replace) << lor_pair(q @ A)
        if q.nvals == 0:
            break
        rv(q.S) << True
    del rv[index]
    return rv


def ancestors(G, source):
    if source not in G._key_to_id:
        raise KeyError(f"The node {source} is not in the graph")
    index = G._key_to_id[source]
    A = G._A
    q = Vector.from_coo(index, True, size=A.nrows, name="q")
    rv = q.dup(name="descendants")
    for _ in range(A.nrows):
        q(~rv.S, replace) << lor_pair(A @ q)
        if q.nvals == 0:
            break
        rv(q.S) << True
    del rv[index]
    return rv



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph

from ..exception import NodeNotFound

__all__ = ["has_path"]


def has_path(G, source, target):
    G = to_graph(G)
    try:
        return algorithms.has_path(G, source, target)
    except KeyError as e:
        raise NodeNotFound(*e.args) from e



from graphblas import Vector
from graphblas.semiring import plus_first, plus_times

from graphblas_algorithms import Graph
from graphblas_algorithms.algorithms._helpers import is_converged
from graphblas_algorithms.algorithms.exceptions import ConvergenceFailure

__all__ = ["pagerank"]


def pagerank(
    G: Graph,
    alpha=0.85,
    personalization=None,
    max_iter=100,
    tol=1e-06,
    nstart=None,
    dangling=None,
    row_degrees=None,
    name="pagerank",
) -> Vector:
    A = G._A
    N = A.nrows
    if A.nvals == 0:
        return Vector(float, N, name=name)

    # Initial vector
    x = Vector(float, N, name="x")
    if nstart is None:
        x[:] = 1.0 / N
    else:
        denom = nstart.reduce().get(0)
        if denom == 0:
            raise ZeroDivisionError("nstart sums to 0")
        x << nstart / denom

    # Personalization vector or scalar
    if personalization is None:
        p = 1.0 / N
    else:
        denom = personalization.reduce().get(0)
        if denom == 0:
            raise ZeroDivisionError("personalization sums to 0")
        p = (personalization / denom).new(name="p")

    # Inverse of row_degrees
    # Fold alpha constant into S
    if row_degrees is None:
        row_degrees = G.get_property("plus_rowwise+")  # XXX: What about self-edges?
    S = (alpha / row_degrees).new(name="S")

    if A.ss.is_iso:
        # Fold iso-value of A into S
        # This lets us use the plus_first semiring, which is faster
        iso_value = A.ss.iso_value
        if iso_value != 1:
            S *= iso_value
        semiring = plus_first[float]
    else:
        semiring = plus_times[float]

    is_dangling = S.nvals < N
    if is_dangling:
        dangling_mask = Vector(float, N, name="dangling_mask")
        dangling_mask(mask=~S.S) << 1.0
        # Fold alpha constant into dangling_weights (or dangling_mask)
        if dangling is not None:
            dangling_weights = (alpha / dangling.reduce().get(0) * dangling).new(
                name="dangling_weights"
            )
        elif personalization is None:
            # Fast case (and common case); is iso-valued
            dangling_mask(mask=dangling_mask.S) << alpha * p
        else:
            dangling_weights = (alpha * p).new(name="dangling_weights")

    # Fold constant into p
    p *= 1 - alpha

    # Power iteration: make up to max_iter iterations
    xprev = Vector(float, N, name="x_prev")
    w = Vector(float, N, name="w")
    for _ in range(max_iter):
        xprev, x = x, xprev

        # x << alpha * ((xprev * S) @ A + "dangling_weights") + (1 - alpha) * p
        x << p
        if is_dangling:
            if dangling is None and personalization is None:
                # Fast case: add a scalar; x is still iso-valued (b/c p is also scalar)
                x += xprev @ dangling_mask
            else:
                # Add a vector
                x += plus_first(xprev @ dangling_mask) * dangling_weights
        w << xprev * S
        x += semiring(w @ A)  # plus_first if A.ss.is_iso else plus_times

        if is_converged(xprev, x, tol):  # sum(abs(xprev - x)) < N * tol
            x.name = name
            return x
    raise ConvergenceFailure(max_iter)



from setuptools import find_packages, setup

import versioneer

extras_require = {
    "test": ["networkx", "pytest", "scipy"],
}
extras_require["complete"] = sorted({v for req in extras_require.values() for v in req})

with open("requirements.txt") as f:
    install_requires = f.read().strip().split("\n")
with open("README.md") as f:
    long_description = f.read()

setup(
    name="graphblas-algorithms",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    entry_points={"networkx.plugins": "graphblas = graphblas_algorithms.interface:Dispatcher"},
    description="Graph algorithms written in GraphBLAS",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jim Kitchen and Erik Welch",
    author_email="erik.n.welch@gmail.com,jim22k@gmail.com",
    url="https://github.com/python-graphblas/graphblas-algorithms",
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=install_requires,
    extras_require=extras_require,
    include_package_data=True,
    license="Apache License 2.0",
    keywords=[
        "graphblas",
        "graph",
        "sparse",
        "matrix",
        "lagraph",
        "suitesparse",
        "Networks",
        "Graph Theory",
        "Mathematics",
        "network",
        "discrete mathematics",
        "math",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    zip_safe=False,
)



from graphblas import monoid
from graphblas.semiring import any_pair, plus_first

from .boundary import edge_boundary, node_boundary

__all__ = [
    "cut_size",
    "volume",
    "normalized_cut_size",
    "conductance",
    "edge_expansion",
    "mixing_expansion",
    "node_expansion",
    "boundary_expansion",
]


def cut_size(G, S, T=None, *, is_weighted=False):
    edges = edge_boundary(G, S, T, is_weighted=is_weighted)
    if is_weighted:
        rv = edges.reduce_scalar(monoid.plus).get(0)
    else:
        rv = edges.nvals
    if G.is_directed():
        edges = edge_boundary(G, T, S, is_weighted=is_weighted)
        if is_weighted:
            rv += edges.reduce_scalar(monoid.plus).get(0)
        else:
            rv += edges.nvals
    return rv


def volume(G, S, *, weighted=False):
    if weighted:
        degrees = plus_first(G._A @ S)
    else:
        degrees = G.get_property("row_degrees+", mask=S.S)
    return degrees.reduce(monoid.plus).get(0)


def normalized_cut_size(G, S, T=None):
    num_cut_edges = cut_size(G, S, T)
    volume_S = volume(G, S)
    volume_T = volume(G, T)
    return num_cut_edges * ((1 / volume_S) + (1 / volume_T))


def conductance(G, S, T=None):
    num_cut_edges = cut_size(G, S, T)
    volume_S = volume(G, S)
    volume_T = volume(G, T)
    return num_cut_edges / min(volume_S, volume_T)


def edge_expansion(G, S, T=None):
    num_cut_edges = cut_size(G, S, T)
    if T is None:
        Tnvals = S.size - S.nvals
    else:
        Tnvals = T.nvals
    return num_cut_edges / min(S.nvals, Tnvals)


def mixing_expansion(G, S, T=None):
    num_cut_edges = cut_size(G, S, T)
    return num_cut_edges / G._A.nvals  # Why no factor of 2 in denominator?


def node_expansion(G, S):
    neighborhood = any_pair(S @ G._A)
    return neighborhood.nvals / S.nvals


def boundary_expansion(G, S):
    result = node_boundary(G, S)
    return result.nvals / S.nvals



from graphblas import monoid

__all__ = ["is_regular", "is_k_regular"]


def is_regular(G):
    if not G.is_directed():
        degrees = G.get_property("degrees+")
        if degrees.nvals != degrees.size:
            return False
        d = degrees.get(0)
        return (degrees == d).reduce(monoid.land).get(True)
    else:
        row_degrees = G.get_property("row_degrees+")
        if row_degrees.nvals != row_degrees.size:
            return False
        column_degrees = G.get_property("column_degrees+")
        if column_degrees.nvals != column_degrees.size:
            return False
        d = row_degrees.get(0)
        if not (row_degrees == d).reduce(monoid.land):
            return False
        d = column_degrees.get(0)
        return (column_degrees == d).reduce(monoid.land).get(True)


def is_k_regular(G, k):
    degrees = G.get_property("degrees+")
    if degrees.nvals != degrees.size:
        return False
    return (degrees == k).reduce(monoid.land).get(True)



import graphblas as gb

from graphblas_algorithms import Graph
from graphblas_algorithms.algorithms import cluster


def test_triangles_full():
    # Including self-edges!
    G = gb.Matrix(bool, 5, 5)
    G[:, :] = True
    G2 = gb.select.offdiag(G).new()
    G = Graph(G)
    G2 = Graph(G2)
    result = cluster.triangles(G)
    expected = gb.Vector(int, 5)
    expected[:] = 6
    assert result.isequal(expected)
    result = cluster.triangles(G2)
    assert result.isequal(expected)
    mask = gb.Vector(bool, 5)
    mask[0] = True
    mask[3] = True
    result = cluster.triangles(G, mask=mask.S)
    expected = gb.Vector(int, 5)
    expected[0] = 6
    expected[3] = 6
    assert result.isequal(expected)
    result = cluster.triangles(G2, mask=mask.S)
    assert result.isequal(expected)
    assert cluster.single_triangle(G, 1) == 6
    assert cluster.single_triangle(G, 0) == 6
    assert cluster.single_triangle(G2, 0) == 6
    assert cluster.total_triangles(G2) == 10
    assert cluster.total_triangles(G) == 10
    assert cluster.transitivity(G) == 1.0
    assert cluster.transitivity(G2) == 1.0
    result = cluster.clustering(G)
    expected = gb.Vector(float, 5)
    expected[:] = 1
    assert result.isequal(expected)
    result = cluster.clustering(G2)
    assert result.isequal(expected)
    assert cluster.single_clustering(G, 0) == 1
    assert cluster.single_clustering(G2, 0) == 1
    expected(mask.S, replace=True) << 1
    result = cluster.clustering(G, mask=mask.S)
    assert result.isequal(expected)
    result = cluster.clustering(G2, mask=mask.S)
    assert result.isequal(expected)
    assert cluster.average_clustering(G) == 1
    assert cluster.average_clustering(G2) == 1
    assert cluster.average_clustering(G, mask=mask.S) == 1
    assert cluster.average_clustering(G2, mask=mask.S) == 1



from graphblas.semiring import plus_pair

__all__ = ["intra_community_edges", "inter_community_edges"]


def intra_community_edges(G, partition):
    A = G._A
    count = 0
    for block in partition:
        # is A or A.T faster?
        count += plus_pair(A @ block).new(mask=block.S).reduce().get(0)
    return count


def inter_community_edges(G, partition):
    A = G._A
    count = 0
    for block in partition:
        # is A or A.T faster?
        count += plus_pair(A @ block).new(mask=~block.S).reduce().get(0)
    return count
    # Alternative implementation if partition is complete set:
    # return G._A.nvals - intra_community_edges_core(G, partition)



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph

__all__ = ["is_simple_path"]


def is_simple_path(G, nodes):
    G = to_graph(G)
    return algorithms.is_simple_path(G, nodes)



from graphblas_algorithms import algorithms
from graphblas_algorithms.classes.digraph import to_graph

from ..exception import ArpackNoConvergence

__all__ = ["hits"]


def hits(G, max_iter=100, tol=1.0e-8, nstart=None, normalized=True):
    G = to_graph(G, weight="weight", dtype=float)
    if len(G) == 0:
        return {}, {}
    x = G.dict_to_vector(nstart, dtype=float, name="nstart")
    try:
        h, a = algorithms.hits(G, max_iter=max_iter, tol=tol, nstart=x, normalized=normalized)
    except algorithms.exceptions.ConvergenceFailure as e:
        if max_iter < 1:
            raise ValueError(*e.args) from e
        else:
            raise ArpackNoConvergence(*e.args, (), ()) from e
        # TODO: it would be nice if networkx raised their own exception, such as:
        # raise nx.PowerIterationFailedConvergence(*e.args) from e
    return G.vector_to_nodemap(h, fillvalue=0), G.vector_to_nodemap(a, fillvalue=0)
